import datetime as dt
import json
import re
from unittest import TestCase
from unittest.mock import patch

import responses
from aioresponses import aioresponses
from flask import url_for
from flask_jwt_extended import create_access_token

from dm.domain.entities import Gate, Locker, Catalog
from dm.domain.entities import Server, Route, Dimension, User
from dm.network.auth import HTTPBearerAuth
from dm.utils.helpers import generate_dimension
from dm.web import create_app, db
from dm.web.background_tasks import TempRoute, update_table_routing_cost, process_catalog_route_table
from tests.helpers import set_callbacks


class TestUpdateTableRoutingCost(TestCase):

    def setUp(self):
        """Create and configure a new app instance for each test."""
        # create the app with common test config
        self.app = create_app('test')
        self.app_context = self.app.app_context()
        self.app_context.push()
        db.create_all()
        User.set_initial()
        self.client = self.app.test_client()

    def tearDown(self) -> None:
        db.session.remove()
        db.drop_all()
        self.app_context.pop()

    @patch('dm.web.background_tasks.check_host', autospec=True)
    @patch('dm.web.background_tasks.ping', autospec=True)
    @responses.activate
    def test_update_table_routing_cost_scenario1(self, mocked_ping, mocked_check_host):
        s1 = Server(id='123e4567-e89b-12d3-a456-426655440001', name='node1')
        g1 = Gate(id='123e4567-e89b-12d3-a456-426655440011', server=s1, port=5001,
                  dns=s1.name)

        s2 = Server(id='123e4567-e89b-12d3-a456-426655440002', name='node2')
        g2 = Gate(id='123e4567-e89b-12d3-a456-426655440012', server=s2, port=5002,
                  dns=s2.name)
        s3 = Server(id='123e4567-e89b-12d3-a456-426655440003', name='node3', me=True)
        g3 = Gate(id='123e4567-e89b-12d3-a456-426655440013', server=s3, port=5003,
                  dns=s3.name)
        Route(s1, gate=g1, cost=0)
        db.session.add_all([s1, s2, s3])

        def callback(request):

            if request.url == Server.query.get('123e4567-e89b-12d3-a456-426655440001').url('api_1_0.routes'):
                msg = {"server_id": '123e4567-e89b-12d3-a456-426655440001',
                       "route_list": [
                           dict(destination_id='123e4567-e89b-12d3-a456-426655440002',
                                gate_id='123e4567-e89b-12d3-a456-426655440012', proxy_server_id=None, cost=0),
                           dict(destination_id='123e4567-e89b-12d3-a456-426655440003',
                                gate_id='123e4567-e89b-12d3-a456-426655440013', proxy_server_id=None, cost=0),
                       ]}
            elif request.url == Server.query.get('123e4567-e89b-12d3-a456-426655440002').url('api_1_0.routes'):
                msg = {"server_id": '123e4567-e89b-12d3-a456-426655440002',
                       "route_list": [
                           dict(destination_id='123e4567-e89b-12d3-a456-426655440001',
                                gate_id='123e4567-e89b-12d3-a456-426655440011', proxy_server_id=None, cost=0),
                       ]}
            else:
                raise
            return 200, {}, json.dumps(msg)

        responses.add_callback(responses.GET, re.compile('^https?://.*$'), callback=callback,
                               content_type='application/json')

        def ping(server, *args, **kwargs):
            if str(server.id) == '123e4567-e89b-12d3-a456-426655440001':
                return 0, None
            elif str(server.id) == '123e4567-e89b-12d3-a456-426655440002':
                return 0, None
            elif str(server.id) == '123e4567-e89b-12d3-a456-426655440003':
                return None, None

        mocked_ping.side_effect = ping

        def check_host(host, port, *args, **kwargs):
            if host == 'node1':
                return True
            elif host == 'node2':
                return True
            elif host == 'node3':
                return True

        mocked_check_host.side_effect = check_host

        changed_routes = update_table_routing_cost(discover_new_neighbours=True, check_current_neighbours=True)

        self.assertEqual(g1, s1.route.gate)
        self.assertEqual(None, s1.route.proxy_server)
        self.assertEqual(0, s1.route.cost)

        self.assertEqual(g2, s2.route.gate)
        self.assertEqual(None, s2.route.proxy_server)
        self.assertEqual(0, s2.route.cost)

        self.assertIsNone(s3.route)
        self.assertDictEqual({s2: (None, g2, 0)}, changed_routes)

    @patch('dm.web.background_tasks.check_host', autospec=True)
    @patch('dm.web.background_tasks.ping', autospec=True)
    @responses.activate
    def test_update_table_routing_cost_scenario2(self, mocked_ping, mocked_check_host):
        s1 = Server(id='123e4567-e89b-12d3-a456-426655440001', name='node1', me=True)
        g1 = Gate(id='123e4567-e89b-12d3-a456-426655440011', server=s1, port=5001,
                  dns=s1.name)

        s2 = Server(id='123e4567-e89b-12d3-a456-426655440002', name='node2')
        g2 = Gate(id='123e4567-e89b-12d3-a456-426655440012', server=s2, port=5002,
                  dns=s2.name)
        s3 = Server(id='123e4567-e89b-12d3-a456-426655440003', name='node3')
        g3 = Gate(id='123e4567-e89b-12d3-a456-426655440013', server=s3, port=5003,
                  dns=s3.name)
        Route(s2, gate=g2, cost=0)
        Route(s3, gate=g3, cost=0)
        db.session.add_all([s1, s2, s3])

        def callback(request):
            if request.url == Server.query.get('123e4567-e89b-12d3-a456-426655440002').url('api_1_0.routes'):
                msg = {"server_id": '123e4567-e89b-12d3-a456-426655440002',
                       "route_list": [
                           dict(destination_id='123e4567-e89b-12d3-a456-426655440001',
                                gate_id='123e4567-e89b-12d3-a456-426655440011', proxy_server_id=None, cost=0),
                           dict(destination_id='123e4567-e89b-12d3-a456-426655440003',
                                gate_id='123e4567-e89b-12d3-a456-426655440013', proxy_server_id=None, cost=0),
                       ]}
            else:
                raise
            return 200, {}, json.dumps(msg)

        responses.add_callback(responses.GET, re.compile('^https?://.*$'), callback=callback,
                               content_type='application/json')

        def ping(server, *args, **kwargs):
            if str(server.id) == '123e4567-e89b-12d3-a456-426655440001':
                return 0, None
            elif str(server.id) == '123e4567-e89b-12d3-a456-426655440002':
                return 0, None
            elif str(server.id) == '123e4567-e89b-12d3-a456-426655440003':
                return None, None

        mocked_ping.side_effect = ping

        def check_host(host, port, *args, **kwargs):
            if host == 'node1':
                return True
            elif host == 'node2':
                return True
            elif host == 'node3':
                return False

        mocked_check_host.side_effect = check_host

        changed_routes = update_table_routing_cost(discover_new_neighbours=True, check_current_neighbours=True)

        self.assertIsNone(s1.route)

        self.assertEqual(g2, s2.route.gate)
        self.assertEqual(None, s2.route.proxy_server)
        self.assertEqual(0, s2.route.cost)

        self.assertEqual(None, s3.route.gate)
        self.assertEqual(s2, s3.route.proxy_server)
        self.assertEqual(1, s3.route.cost)

        self.assertDictEqual({s3: (s2, None, 1)}, changed_routes)

    @patch('dm.web.background_tasks.check_host', autospec=True)
    @patch('dm.web.background_tasks.ping', autospec=True)
    @responses.activate
    def test_update_table_routing_cost_scenario3(self, mocked_ping, mocked_check_host):
        # Node 1 loses connection to gate's Node 2 and sets the second gate as default gate
        s1 = Server(id='123e4567-e89b-12d3-a456-426655440001', name='node1', me=True)
        g1 = Gate(id='123e4567-e89b-12d3-a456-426655440011', server=s1, port=5001,
                  dns=s1.name)

        s2 = Server(id='123e4567-e89b-12d3-a456-426655440002', name='node2')
        g21 = Gate(id='123e4567-e89b-12d3-a456-426655440012', server=s2, port=5012,
                   dns=s2.name)
        g22 = Gate(id='123e4567-e89b-12d3-a456-426655440022', server=s2, port=5022,
                   dns=s2.name)

        Route(s2, gate=g21, cost=0)
        db.session.add_all([s1, s2])

        def callback(request):
            if request.url == Server.query.get('123e4567-e89b-12d3-a456-426655440002').url('api_1_0.routes').replace(
                    '5012', '5022'):
                msg = {"server_id": '123e4567-e89b-12d3-a456-426655440002',
                       "route_list": [
                           dict(destination_id='123e4567-e89b-12d3-a456-426655440001',
                                gate_id='123e4567-e89b-12d3-a456-426655440011', proxy_server_id=None, cost=0),
                       ]}
            else:
                raise
            return 200, {}, json.dumps(msg)

        responses.add_callback(responses.GET, re.compile('^https?://.*$'), callback=callback,
                               content_type='application/json')

        def ping(server, *args, **kwargs):
            if str(server.id) == '123e4567-e89b-12d3-a456-426655440001':
                return 0, None
            elif str(server.id) == '123e4567-e89b-12d3-a456-426655440002':
                return None, None

        mocked_ping.side_effect = ping

        def check_host(host, port, *args, **kwargs):
            if host == 'node1' and port == 5001:
                return True
            elif host == 'node2' and port == 5012:
                return False
            elif host == 'node2' and port == 5022:
                return True
            else:
                return False

        mocked_check_host.side_effect = check_host

        changed_routes = update_table_routing_cost(discover_new_neighbours=True, check_current_neighbours=True)

        self.assertIsNone(s1.route)

        self.assertEqual(g22, s2.route.gate)
        self.assertEqual(None, s2.route.proxy_server)
        self.assertEqual(0, s2.route.cost)

        self.assertDictEqual({s2: TempRoute(
            None, g22, 0)}, changed_routes)

    @patch('dm.web.background_tasks.check_host', autospec=True)
    @patch('dm.web.background_tasks.ping', autospec=True)
    @responses.activate
    def test_update_table_routing_cost_scenario4(self, mocked_ping, mocked_check_host):
        s1 = Server(id='123e4567-e89b-12d3-a456-426655440001', name='node1', me=True)
        g1 = Gate(id='123e4567-e89b-12d3-a456-426655440011', server=s1, port=5001,
                  dns=s1.name)

        s2 = Server(id='123e4567-e89b-12d3-a456-426655440002', name='node2')
        g2 = Gate(id='123e4567-e89b-12d3-a456-426655440012', server=s2, port=5002,
                  dns=s2.name)
        s3 = Server(id='123e4567-e89b-12d3-a456-426655440003', name='node3')
        g3 = Gate(id='123e4567-e89b-12d3-a456-426655440013', server=s3, port=5003,
                  dns=s3.name)
        Route(s2, gate=g2, cost=0)
        Route(s3, gate=g3, cost=0)
        db.session.add_all([s1, s2, s3])

        def callback(request):
            if request.url == Server.query.get('123e4567-e89b-12d3-a456-426655440002').url('api_1_0.routes'):
                msg = {"server_id": '123e4567-e89b-12d3-a456-426655440002',
                       "route_list": [
                           dict(destination_id='123e4567-e89b-12d3-a456-426655440001',
                                gate_id='123e4567-e89b-12d3-a456-426655440011', proxy_server_id=None, cost=0),
                           dict(destination_id='123e4567-e89b-12d3-a456-426655440003',
                                gate_id=None, proxy_server_id=None, cost=None),
                       ]}
            else:
                raise
            return 200, {}, json.dumps(msg)

        responses.add_callback(responses.GET, re.compile('^https?://.*$'), callback=callback,
                               content_type='application/json')

        def ping(server, *args, **kwargs):
            if str(server.id) == '123e4567-e89b-12d3-a456-426655440001':
                return 0, None
            elif str(server.id) == '123e4567-e89b-12d3-a456-426655440002':
                return 0, None
            elif str(server.id) == '123e4567-e89b-12d3-a456-426655440003':
                return None, None

        mocked_ping.side_effect = ping

        def check_host(host, port, *args, **kwargs):
            if host == 'node1':
                return True
            elif host == 'node2':
                return True
            elif host == 'node3':
                return False

        mocked_check_host.side_effect = check_host

        changed_routes = update_table_routing_cost(discover_new_neighbours=True, check_current_neighbours=True)

        self.assertIsNone(s1.route)

        self.assertEqual(g2, s2.route.gate)
        self.assertEqual(None, s2.route.proxy_server)
        self.assertEqual(0, s2.route.cost)

        self.assertEqual(None, s3.route.gate)
        self.assertEqual(None, s3.route.proxy_server)
        self.assertEqual(None, s3.route.cost)

        self.assertDictEqual({s3: TempRoute(
            None, None, None)}, changed_routes)

    @patch('dm.web.background_tasks.check_host', autospec=True)
    @patch('dm.web.background_tasks.ping', autospec=True)
    @responses.activate
    def test_update_table_routing_cost_scenario5(self, mocked_ping, mocked_check_host):
        s1 = Server(id='123e4567-e89b-12d3-a456-426655440001', name='node1', me=True)
        g1 = Gate(id='123e4567-e89b-12d3-a456-426655440011', server=s1, port=5001,
                  dns=s1.name)

        s2 = Server(id='123e4567-e89b-12d3-a456-426655440002', name='node2')
        g2 = Gate(id='123e4567-e89b-12d3-a456-426655440012', server=s2, port=5012,
                  dns=s2.name)

        Route(s2, gate=g2, cost=0)
        db.session.add_all([s1, s2])

        def callback(request):
            if request.url == Server.query.get('123e4567-e89b-12d3-a456-426655440002').url('api_1_0.routes'):
                msg = {"server_id": '123e4567-e89b-12d3-a456-426655440002',
                       "route_list": [
                           dict(destination_id='123e4567-e89b-12d3-a456-426655440001',
                                gate_id='123e4567-e89b-12d3-a456-426655440011', proxy_server_id=None, cost=0),
                           dict(destination_id='123e4567-e89b-12d3-a456-426655440003',
                                gate_id='123e4567-e89b-12d3-a456-426655440013', proxy_server_id=None, cost=0),
                       ]}
            else:
                raise
            return 200, {}, json.dumps(msg)

        responses.add_callback(responses.GET, re.compile('^https?://.*$'), callback=callback,
                               content_type='application/json')

        def ping(server, *args, **kwargs):
            if str(server.id) == '123e4567-e89b-12d3-a456-426655440001':
                return 0, None
            elif str(server.id) == '123e4567-e89b-12d3-a456-426655440002':
                return 0, None

        mocked_ping.side_effect = ping

        mocked_check_host.assert_not_called()

        changed_routes = update_table_routing_cost(discover_new_neighbours=True, check_current_neighbours=True)

        self.assertIsNone(s1.route)

        self.assertEqual(g2, s2.route.gate)
        self.assertEqual(None, s2.route.proxy_server)
        self.assertEqual(0, s2.route.cost)

        self.assertDictEqual({}, changed_routes)

    @patch('dm.web.background_tasks.check_host', autospec=True)
    @patch('dm.web.background_tasks.ping', autospec=True)
    @responses.activate
    def test_update_table_routing_cost_scenario6(self, mocked_ping, mocked_check_host):
        # Nodes have localhost and node2 is not a neighbour anymore
        s1 = Server(id='123e4567-e89b-12d3-a456-426655440001', name='node1', me=True)
        g11 = Gate(id='123e4567-e89b-12d3-a456-426655440011', server=s1, port=5000,
                   ip='127.0.0.1')
        g12 = Gate(id='123e4567-e89b-12d3-a456-426655440012', server=s1, port=5000,
                   ip='10.0.0.1')

        s2 = Server(id='123e4567-e89b-12d3-a456-426655440002', name='node2')
        g21 = Gate(id='123e4567-e89b-12d3-a456-426655440021', server=s2, port=5000,
                   ip='127.0.0.1')
        g22 = Gate(id='123e4567-e89b-12d3-a456-426655440022', server=s2, port=5000,
                   ip='10.0.0.2')

        Route(s2, gate=g22, cost=0)
        db.session.add_all([s1, s2])
        db.session.commit()

        def callback(request):
            if request.url == Server.query.get('123e4567-e89b-12d3-a456-426655440001').url('api_1_0.routes'):
                msg = {"server_id": '123e4567-e89b-12d3-a456-426655440001',
                       "route_list": [
                           dict(destination_id='123e4567-e89b-12d3-a456-426655440002',
                                gate_id='123e4567-e89b-12d3-a456-426655440022', proxy_server_id=None, cost=0),
                       ]}
            else:
                raise
            return 200, {}, json.dumps(msg)

        responses.add(responses.GET, url=Server.query.get('123e4567-e89b-12d3-a456-426655440001').url('api_1_0.routes'),
                      json={"server_id": '123e4567-e89b-12d3-a456-426655440001',
                            "route_list": [
                                dict(destination_id='123e4567-e89b-12d3-a456-426655440002',
                                     gate_id='123e4567-e89b-12d3-a456-426655440022', proxy_server_id=None, cost=0),
                            ]})

        def ping(server, *args, **kwargs):
            if str(server.id) == '123e4567-e89b-12d3-a456-426655440001':
                return 0, None
            elif str(server.id) == '123e4567-e89b-12d3-a456-426655440002':
                return None, None

        mocked_ping.side_effect = ping

        def check_host(host, port, *args, **kwargs):
            if host == '127.0.0.1':
                return True
            elif host == '10.0.0.1':
                return True
            elif host == '10.0.0.2':
                return False
            else:
                raise ConnectionError

        mocked_check_host.side_effect = check_host

        changed_routes = update_table_routing_cost(discover_new_neighbours=True, check_current_neighbours=True)

        self.assertIsNone(s1.route)

        self.assertEqual(None, s2.route.gate)
        self.assertEqual(None, s2.route.proxy_server)
        self.assertEqual(None, s2.route.cost)

        self.assertDictEqual({s2: TempRoute(
            None, None, None)}, changed_routes)

    @patch('dm.web.background_tasks.check_host', autospec=True)
    @patch('dm.web.background_tasks.ping', autospec=True)
    @responses.activate
    def test_update_table_routing_cost_scenario6(self, mocked_ping, mocked_check_host):
        # Node have localhost and node2 appears as a new neighbour
        s1 = Server(id='123e4567-e89b-12d3-a456-426655440001', name='node1', me=True)
        g11 = Gate(id='123e4567-e89b-12d3-a456-426655440011', server=s1, port=5000,
                   ip='127.0.0.1')
        g12 = Gate(id='123e4567-e89b-12d3-a456-426655440012', server=s1, port=5000,
                   ip='10.0.0.1')

        s2 = Server(id='123e4567-e89b-12d3-a456-426655440002', name='node2')
        g21 = Gate(id='123e4567-e89b-12d3-a456-426655440021', server=s2, port=5000,
                   ip='127.0.0.1')
        g22 = Gate(id='123e4567-e89b-12d3-a456-426655440022', server=s2, port=5000,
                   ip='10.0.0.2')

        Route(s2, gate=None, cost=None)
        db.session.add_all([s1, s2])
        db.session.commit()

        def callback(request):
            if re.search("^https?://(10\.0\.0\.1|127\.0\.0\.1):5000" + url_for('api_1_0.routes', _external=False),
                         request.url):
                msg = {"server_id": '123e4567-e89b-12d3-a456-426655440001',
                       "route_list": [
                           dict(destination_id='123e4567-e89b-12d3-a456-426655440002',
                                gate_id='123e4567-e89b-12d3-a456-426655440022', proxy_server_id=None, cost=0),
                       ]}
            elif re.search("^https?://(10\.0\.0\.2):5000" + url_for('api_1_0.routes', _external=False),
                           request.url):
                msg = {"server_id": '123e4567-e89b-12d3-a456-426655440002',
                       "route_list": [
                           dict(destination_id='123e4567-e89b-12d3-a456-426655440001',
                                gate_id='123e4567-e89b-12d3-a456-426655440012', proxy_server_id=None, cost=0),
                       ]}
            else:
                raise ConnectionError()
            return 200, {}, json.dumps(msg)

        responses.add_callback(responses.GET, re.compile('^https?://.*$'), callback=callback,
                               content_type='application/json')

        def ping(server, *args, **kwargs):
            if str(server.id) == '123e4567-e89b-12d3-a456-426655440001':
                return 0, None
            elif str(server.id) == '123e4567-e89b-12d3-a456-426655440002':
                return None, None

        mocked_ping.side_effect = ping

        def check_host(host, port, *args, **kwargs):
            if host == '127.0.0.1':
                return True
            elif host == '10.0.0.1':
                return True
            elif host == '10.0.0.2':
                return True
            else:
                raise ConnectionError

        mocked_check_host.side_effect = check_host

        changed_routes = update_table_routing_cost(discover_new_neighbours=True, check_current_neighbours=True)

        self.assertIsNone(s1.route)

        self.assertEqual(g22, s2.route.gate)
        self.assertEqual(None, s2.route.proxy_server)
        self.assertEqual(0, s2.route.cost)

        self.assertDictEqual({s2: TempRoute(
            None, g22, 0)}, changed_routes)

        db.session.commit()


class TestProcessCatalogRouteTable(TestCase):

    @patch('dm.domain.entities.get_now')
    def setUp(self, mock_now):
        """Create and configure a new app instance for each test."""
        # create the app with common test config
        self.app = create_app('test')
        self.app.config['SECURIZER'] = True
        mock_now.return_value = dt.datetime(2019, 4, 1, tzinfo=dt.timezone.utc)
        with self.app.app_context():
            self.dim = generate_dimension('test')
            self.dim.current = True
            self.json_dim = self.dim.to_json()
            self.client = self.app.test_client()

            db.create_all()
            Locker.set_initial()
            User.set_initial()

            self.auth = HTTPBearerAuth(create_access_token(User.get_by_user('root').id))

            server = Server('node1', dns_or_ip='127.0.0.1', port=8000, me=True)
            for g in server.gates:
                g.last_modified_at = dt.datetime(2019, 4, 1, tzinfo=dt.timezone.utc)
            db.session.add_all([server, self.dim])
            db.session.commit()

            # dump data
            self.json_node1 = server.to_json(add_gates=True)
            self.json_users = [u.to_json() for u in User.query.all()]

        self.app2 = create_app('test')
        self.app2.config['SECURIZER'] = True
        self.client2 = self.app2.test_client()
        mock_now.return_value = dt.datetime(2019, 4, 2, tzinfo=dt.timezone.utc)
        with self.app2.app_context():
            db.create_all()
            Locker.set_initial()
            User.set_initial()

            me = Server('node2', port=8000, me=True, granules='granule',
                        last_modified_at=dt.datetime(2019, 4, 2, tzinfo=dt.timezone.utc))
            for g in me.gates:
                g.last_modified_at = dt.datetime(2019, 4, 2, tzinfo=dt.timezone.utc)
            db.session.add(me)

            src_server = Server.from_json(self.json_node1)
            Route(src_server, cost=0)
            db.session.add(src_server)

            dim = Dimension.from_json(self.json_dim)
            dim.current = True
            db.session.add(dim)

            users = [User.from_json(ju) for ju in self.json_users]
            db.session.add_all(users)

            db.session.commit()

            # dump data
            self.json_node2 = me.to_json(add_gates=True)

        mock_now.return_value = dt.datetime(2019, 4, 1, tzinfo=dt.timezone.utc)
        with self.app.app_context():
            node2 = Server.from_json(self.json_node2)
            node2.last_modified_at = dt.datetime(2019, 4, 1, tzinfo=dt.timezone.utc)
            Route(node2, cost=0)
            db.session.add(node2)
            db.session.commit()

    def tearDown(self) -> None:
        with self.app2.app_context():
            db.session.remove()
            db.drop_all()

        with self.app.app_context():
            db.session.remove()
            db.drop_all()


    @responses.activate
    @aioresponses()
    @patch('dm.web.background_tasks.update_table_routing_cost', return_value=True)
    @patch('dm.web.background_tasks.upgrade_version', return_value=False)
    def test_catalog(self, m, mock_version, mock_routing):
        # test all system from process_catalog_route_table to lock server and upgrade catalog
        set_callbacks([("(127.0.0.1|node1)", self.app.test_client()),
                       ("node2", self.app2.test_client())], m=m)

        with self.app.app_context():
            datemark = Catalog.max_catalog()
        self.assertEqual(dt.datetime(2019, 4, 1, tzinfo=dt.timezone.utc), datemark)

        process_catalog_route_table(self.app)

        with self.app.app_context():
            datemark = Catalog.max_catalog()
        self.assertEqual(dt.datetime(2019, 4, 2, tzinfo=dt.timezone.utc), datemark.astimezone(dt.timezone.utc))



"""
Usage: dshell orch list [--id ID|--name NAME|--like LIKE] [--version N] [--last N]
       dshell orch create (- | FILE)
       dshell orch run ID [[--parameter PARAM=VALUE]... | [--json-parameters JSON]]
                          ((--target NAME=VALUE)... | --json-target JSON)
                          [--background]

Command to interact with orchestration

Arguments:
    FILE                     File which has the orchestration to create
    ID                       Orchestration ID to launch
    NAME                     Orchestration's name

Options:
    --id ID                  Action template id to show
    --last N                 Shows last N servers
    --like LIKE              List action templates whose names contain LIKE
    --json-parameters JSON   Json parameters. Ex.: '{"folder":"/tmp",
    --json-target JSON       Target in json format. Ex.: '{"all":["node1", "node2"], "backend": "granule-backend"}'
    --name NAME              Action templates' name to list
    --parameter PARAM=VALUE  Parameter passed to run orchestration. Examples: --parameter folder=/tmp
    --target NAME=VALUE      Target to run the orchestrations NAME is the target name and value the granules or hosts
                             select. Example: --target all=node1,node2 --target backend=granule-backend
    --version N              Versions to list
    --background             Runs the orchestration in background
"""
import json
import sys
from pprint import pprint

from docopt import docopt
from schema import Schema, Or, Use, And, SchemaError

import dm.dshell.network as ntwrk
from dm.dshell.commands import orch_list, orch_run

orch_schema = Schema(
    {
        'FILE': Or(None, Use(open, error="Unable to open file. File not found")),
        '--last': Or(None, And(Use(int, error="Not a valid integer"), And(lambda n: 0 < n,
                                                                          error="integer must be above 0"))),
        str: object}, )


def main(args):
    argv = docopt(__doc__, args)

    try:
        argv = orch_schema.validate(argv)
    except SchemaError as e:
        exit(str(e))
    if argv['list']:
        orch_list(iden=argv['--id'], name=argv['--name'], version=argv['--version'], like=argv['--like'],
                    last=argv['--last'])
    elif argv['create']:
        data = None
        if argv['-']:
            content = sys.stdin.read()
            part = json.loads(content)
            data = part if isinstance(part, list) else [part]
        elif argv['FILE']:
            data = []
            content = argv['FILE'].read()
            part = json.loads(content)
            data.extend(part) if isinstance(part, list) else data.append(part)
        else:
            exit("No file specified")
        if data:
            resp = ntwrk.post(f'api_1_0.orchestrationlist', json=data)
            if resp.ok:
                if resp.msg:
                    pprint(resp.msg)
            else:
                pprint(resp.msg if resp.code is not None else str(resp.exception) if str(
                    resp.exception) else resp.exception.__class__.__name__)
        else:
            print("No data found to create action template")
    elif argv['run']:
        orchestration_id = argv['ID']

        if argv['--json-parameters'] is None:
            parameters = {}
            for param in argv['PARAM=VALUE']:
                key, value = param.split('=', 1)
                if key not in parameters:
                    parameters[key] = value
                else:
                    if isinstance(parameters[key], list):
                        parameters[key].append(value)
                    else:
                        parameters[key] = [parameters[key], value]
        else:
            parameters = json.loads(argv['--json-parameters'])

        if argv['--json-target'] is None:
            target = {}
            for param in argv['NAME=VALUE']:
                key, value = param.split('=', 1)
                if ',' in value:
                    target[key] = value.split(',')
                else:
                    target[key] = value
        else:
            target = json.loads(argv['--json-target'])

        orch_run(orchestration_id, params=parameters, hosts=target, background=argv['--background'] or False)



import argparse
from unittest import TestCase

from dm.dshell.argparse_raise import create_parser, GuessArgumentParser


class TestArgparseRaise(TestCase):
    def test_create_parser(self):
        def status(args):
            pass

        def server_list(args):
            pass

        data = {
            'status': [{'argument': 'node', 'completer': None}, status],
            'server': {
                'list': [{'argument': '--json', 'action': 'store_true', 'required': False},
                         [{'argument': '--like'},
                          {'argument': '--name', 'completer': None},
                          {'argument': '--id'},
                          {'argument': '--last', 'action': 'store', 'type': int}],
                         server_list
                         ],

            }}

        parser = create_parser(data)
        # namespace = parser.parse_args('status node1'.split())
        # self.assertEqual(status, namespace.func)
        # self.assertEqual('node1', namespace.node)

        namespace = parser.parse_args('server list --j'.split())
        self.assertEqual(server_list, namespace.func)
        self.assertEqual('node1', namespace.name)
        self.assertEqual(True, namespace.json)

        namespace = parser.parse_args('server list --json --name node1'.split())
        self.assertEqual(server_list, namespace.func)
        self.assertEqual('node1', namespace.name)
        self.assertEqual(True, namespace.json)

        with self.assertRaises(argparse.ArgumentError):
            parser.parse_args('server list --name node1 --id 1'.split())


class TestGuessArgparse(TestCase):
    def test_create_parser(self):
        def status(args):
            pass

        def server_list(args):
            pass

        data = {
            'status': [{'argument': 'node', 'completer': None}, status],
            'server': {
                'list': [{'argument': '--json', 'action': 'store_true', 'required': False},
                         [{'argument': '--like'},
                          {'argument': '--name', 'completer': None},
                          {'argument': '--id'},
                          {'argument': '--last', 'action': 'store', 'type': int}],
                         server_list
                         ],

            }}

        parser = create_parser(data, GuessArgumentParser())

        namespace = parser.parse_args('server list --json --name node1 --data aa'.split())
        args = dict(namespace._get_kwargs())

        self.assertNotIn('--data', args)

        namespace = parser.parse_args('server list --json --name node1 --last x'.split())
        args = dict(namespace._get_kwargs())
        self.assertEqual('x', namespace.last)

        namespace = parser.parse_args('server list --j'.split())
        self.assertFalseIn(namespace.json)

    def test_parameters(self):
        data = {
            'status': [{'argument': 'arg1', 'completer': None, 'nargs': 2},
                       {'argument': 'arg2', 'completer': None}],
        }

        parser = create_parser(data, GuessArgumentParser())

        namespace = parser.parse_args('status first'.split())
        args = dict(namespace._get_kwargs())
        self.assertDictEqual({'arg1': ['first'], 'arg2': None}, args)


from unittest import TestCase

from flask_jwt_extended import create_access_token

from dm import defaults
from dm.domain.entities import ActionTemplate, ActionType
from dm.web import create_app, db


class TestApi(TestCase):
    def setUp(self):
        """Create and configure a new app instance for each test."""
        # create the app with common test config
        self.app = create_app('test')
        self.app_context = self.app.app_context()
        self.app_context.push()
        self.client = self.app.test_client()
        self.headers = {"Authorization": f"Bearer {create_access_token('00000000-0000-0000-0000-000000000001')}"}

        db.create_all()
        db.session.commit()

    def tearDown(self) -> None:
        db.session.remove()
        db.drop_all()
        self.app_context.pop()

    def test_to_from_json(self):
        now = defaults.INITIAL_DATEMARK
        at = ActionTemplate(id='aaaaaaaa-1234-5678-1234-56781234aaa1', name='ActionTest2', version=1,
                            action_type=ActionType.ORCHESTRATION,
                            code='test code',
                            parameters={'dir': '/home'},
                            last_modified_at=now)

        db.session.add(at)

        at_json = at.to_json()

        self.assertDictEqual(dict(id='aaaaaaaa-1234-5678-1234-56781234aaa1', name='ActionTest2', version=1,
                                  action_type='ORCHESTRATION', code='test code', parameters={'dir': '/home'},
                                  last_modified_at=now.strftime(defaults.DATEMARK_FORMAT)), at_json)

        at_json['name'] = "ChangedAction"

        smashed = ActionTemplate.from_json(at_json)

        self.assertEqual(at.id, smashed.id)
        self.assertIsNotNone(smashed.id)
        self.assertEqual("ChangedAction", smashed.name)
        self.assertIsNotNone(smashed.name)
        self.assertEqual(at.version, smashed.version)
        self.assertIsNotNone(smashed.version)
        self.assertEqual(at.action_type, smashed.action_type)
        self.assertIsNotNone(smashed.action_type)
        self.assertEqual(at.code, smashed.code)
        self.assertIsNotNone(smashed.code)
        self.assertEqual(at.last_modified_at, smashed.last_modified_at)
        self.assertIsNotNone(smashed.last_modified_at)

        db.session.commit()

        del at
        del smashed

        at = ActionTemplate.query.get(at_json['id'])
        self.assertEqual("ChangedAction", at.name)



"""
Exceptions for Domain Errors
IDs from 2000 to 2099
"""


class GatewayError(Exception):
    """Base Class for Gateway Exceptions"""


class TimeoutError(GatewayError):
    """Timeout reached while trying to execute the process"""


class UnpackError(GatewayError):
    """Error raised when unpacking message"""


class NotValidMessage(UnpackError):
    """Message is not packed"""


class UnknownMessageType(GatewayError):
    """an unknown message type was received"""


class UnknownFunctionMediator(GatewayError):
    """an unknown function was tried to be executed"""



PROXIED_OBJECT = '__proxied_object'


def str2bool(v):
    return str(v).lower() in ("yes", "true", "t", "1")


class InstanceProxy(object):

    def __init__(self, proxied_obj):
        self._self = proxied_obj

    @property
    def _self(self):
        try:
            return object.__getattribute__(self, PROXIED_OBJECT)
        except AttributeError:
            return None

    @_self.setter
    def _self(self, proxied_obj):
        object.__setattr__(self, PROXIED_OBJECT, proxied_obj)
        return self

    def __getattribute__(self, attr):
        super_cls_dict = InstanceProxy.__dict__
        cls_dict = object.__getattribute__(self, '__class__').__dict__
        inst_dict = object.__getattribute__(self, '__dict__')
        if attr in cls_dict or attr in inst_dict or attr in super_cls_dict:
            return object.__getattribute__(self, attr)
        target_obj = object.__getattribute__(self, PROXIED_OBJECT)
        return object.__getattribute__(target_obj, attr)

    def __repr__(self):
        class_name = object.__getattribute__(self, '__class__').__name__
        target_repr = repr(self._self)
        return '<%s( %s )>' % (class_name, target_repr)



import logging
import threading
from contextlib import contextmanager

from sqlalchemy import event, inspect
from sqlalchemy.orm import sessionmaker

from dm.utils.helpers import get_distributed_entities, get_now
from dm.web import db
from .action_template import ActionTemplate, ActionType
from .catalog import Catalog
from .dimension import Dimension
from .execution import StepExecution, OrchExecution
from .gate import Gate
from .locker import Locker, State, Scope
from .log import Log, Mode as LogMode
from .orchestration import Orchestration
from .route import Route
from .server import Server
from .service import Service
from .software import Software, SoftwareServerAssociation
from .step import Step
from .transfer import Transfer, Status as TransferStatus
from .user import User

logger = logging.getLogger('dm.catalog')

__all__ = [
    "ActionTemplate",
    "ActionType",
    "Catalog",
    "Dimension",
    "StepExecution",
    "OrchExecution",
    "Gate",
    "Locker",
    "State",
    "Scope",
    "Log",
    "LogMode",
    "Orchestration",
    "Route",
    "Service",
    "Server",
    "Software",
    "SoftwareServerAssociation",
    "Step",
    "Transfer",
    "TransferStatus",
    "User",
]

catalog = threading.local()


@contextmanager
def bypass_datamark_update():
    db.session.flush()
    update_datemark(False)
    try:
        yield None
    finally:
        db.session.flush()
        update_datemark(True)



def update_datemark(set):
    catalog.datemark = set


def set_events():
    for name, entity in get_distributed_entities():
        def receive_before_insert(mapper, connection, target):
            if not hasattr(catalog, 'data'):
                catalog.data = {}
            if getattr(catalog, 'datemark', True):
                target.last_modified_at = get_now()
            if not target.__class__ in catalog.data:
                catalog.data.update({target.__class__: target.last_modified_at})
            else:
                catalog.data.update(
                    {target.__class__: max(target.last_modified_at, catalog.data[target.__class__])})

        def receive_before_update(mapper, connection, target):
            # validate if target is updated
            it = inspect(target)
            columns = mapper.columns.keys()
            changed = any(
                [attr.history.has_changes() for attr in it.attrs if attr.key in columns])
            if changed:
                if not hasattr(catalog, 'data'):
                    catalog.data = {}
                if getattr(catalog, 'datemark', True):
                    target.last_modified_at = get_now()
                if not target.__class__ in catalog.data:
                    catalog.data.update({target.__class__: target.last_modified_at})
                else:
                    catalog.data.update(
                        {target.__class__: max(target.last_modified_at, catalog.data[target.__class__])})

        event.listen(entity, 'before_insert', receive_before_insert, propagate=False)
        event.listen(entity, 'before_update', receive_before_update, propagate=False)


set_events()


@event.listens_for(db.session, 'after_commit')
def receive_after_commit(session):
    # TODO: run this peace of code in a thread to allow the request end while executing the queries
    if hasattr(catalog, 'data'):
        # create a different session as the request session is already commited
        engine = db.get_engine()
        Session = sessionmaker(bind=engine)
        s = Session()
        for e, last_modified_at in catalog.data.items():
            # last_modified_at = s.query(func.max(e.last_modified_at)).scalar()
            c = s.query(Catalog).filter_by(entity=e.__name__).first()
            if c is None:
                c = Catalog(entity=e.__name__, last_modified_at=last_modified_at)
            else:
                if c.last_modified_at < last_modified_at:
                    c.last_modified_at = last_modified_at
            logger.debug(f'changed catalog {c}')
            s.add(c)
            s.commit()
            del c
        catalog.data = {}
        s.close()


@event.listens_for(Orchestration, 'refresh')
def receive_refresh(target, context, attrs):
    "listen for the 'refresh' event"
    target.init_on_load()



import asyncio
import concurrent
import logging
import typing as t

import aiohttp
import requests
import rsa
from aiohttp import ContentTypeError
from flask import current_app as __ca, current_app, url_for
from requests.exceptions import Timeout

from dm import defaults
from dm.domain.entities import Server, Dimension, Gate
from dm.network.encryptation import pack_msg as _pack_msg, unpack_msg as _unpack_msg
from dm.network.exceptions import NotValidMessage
from dm.utils.helpers import get_now
from dm.utils.typos import Kwargs, tJSON
from dm.web import errors

requests.packages.urllib3.disable_warnings()

logger = logging.getLogger('dm.network')


class Response:

    def __init__(self, msg: tJSON = None, code: int = None, exception: Exception = None,
                 server=None,
                 url=None, headers=None):
        self.code = code
        self.msg = msg
        self.exception = exception
        self.server = server
        self.url = url
        self.headers = headers

    def __eq__(self, other):
        return isinstance(other, self.__class__) \
               and self.code == other.code \
               and self.msg == other.msg \
               and self.server == other.server \
               and self.url == other.url \
               and ((self.exception.__class__ == other.exception.__class__ and
                     self.exception.args == other.exception.args) if self.exception else True)

    # backward compatibility <= v0.1-b10
    def __iter__(self):
        yield self.exception or self.msg
        yield self.code

    # backward compatibility <= v0.1-b10
    def __getitem__(self, item):
        if item == 0:
            return self.exception or self.msg
        elif item == 1:
            return self.code
        raise KeyError(item)

    def __repr__(self):
        return f"Response(server={self.server}, url={self.url}, code={self.code})"

    def __str__(self):
        if self.exception:
            return f"{self.exception.__class__.__name__}: {self.exception}"
        return f"{self.code}, {self.msg}"

    def to_dict(self):
        dump = dict()
        if self.code:
            dump.update(code=self.code, response=self.msg)
            if self.url:
                dump.update(url=self.url)
            if self.server:
                dump.update(server=dict(id=str(self.server.id), name=self.server.name))
        else:
            dump.update(exception=str(self.exception) if str(self.exception) else self.exception.__class__.__name__)
        return dump

    def raise_on_error(self):
        if self.exception:
            raise self.exception

    def raise_for_status(self):
        if self.code < 200 or 299 < self.code:
            raise errors.HTTPError(self)

    def raise_if_not_ok(self):
        self.raise_on_error()
        self.raise_for_status()

    @property
    def ok(self):
        return not bool(self.exception) and bool(self.code) and 200 <= self.code <= 299



def pack_msg(data, *args, **kwargs):
    if not __ca.config['SECURIZER']:
        return data
    else:
        # if not ('symmetric_key' in kwargs or 'cipher_key' in kwargs):
        #     try:
        #         kwargs['cipher_key'] = session.get('cipher_key')
        #     except RuntimeError:
        #         pass
        # if generate_key:
        #     kwargs.pop('symmetric_key', None)
        #     kwargs.pop('cipher_key', None)
        dim = None
        if dim is None:
            dim = Dimension.get_current()
            if dim is None:
                raise ValueError('No dimension found but SECURIZER set')
        return _pack_msg(data, *args, source=Server.get_current(),
                         pub_key=dim.public,
                         priv_key=dim.private,
                         **kwargs)


def pack_msg2(data, *args, **kwargs):
    if not __ca.config['SECURIZER']:
        return data
    else:
        return _pack_msg(data, *args, **kwargs)


def unpack_msg(data, *args, **kwargs):
    if not __ca.config['SECURIZER']:
        return data
    else:
        # if 'key' in data:
        #     cipher_key = base64.b64decode(data.get('key'))
        # else:
        #     try:
        #         cipher_key = session.get('cipher_key', None)
        #     except RuntimeError:
        #         cipher_key = None
        # try:
        #     session['cipher_key'] = cipher_key
        # except RuntimeError:
        #     pass
        if not 'error' in data:
            dim = None
            if dim is None:
                dim = Dimension.get_current()
                if dim is None:
                    raise ValueError('No dimension found but SECURIZER set')
            return _unpack_msg(data, *args, pub_key=dim.public,
                               priv_key=dim.private, **kwargs)
        else:
            return data


def unpack_msg_no_ctx(data, securizer: bool, public_key: rsa.PublicKey, private_key: rsa.PrivateKey, *args, **kwargs):
    if not securizer:
        return data
    else:
        if not 'error' in data:
            return _unpack_msg(data, *args, pub_key=public_key,
                               priv_key=private_key, **kwargs)
        else:
            return data


def unpack_msg2(data, *args, **kwargs):
    if not __ca.config['SECURIZER']:
        return data
    else:
        if not 'error' in data:
            return _unpack_msg(data, *args, **kwargs)
        else:
            return data


def ping(dest: t.Union[Server, Gate], source: Server, retries=3, timeout=3, verify=False):
    tries = 0
    cost = None
    elapsed = None
    exc = None

    if isinstance(dest, Gate):
        server = dest.server
        schema = current_app.config['PREFERRED_URL_SCHEME'] or 'https'
        url = f"{schema}://{dest}/{url_for('root.ping', _external=False)}"
    else:
        server = dest
        try:
            url = dest.url('root.ping')
        except:
            logger.exception(f"exception while getting url to {dest}")
            return None, None
    while tries < retries:
        try:
            tries += 1
            resp = requests.post(url,
                                 json={'start_time': get_now().strftime(defaults.DATETIME_FORMAT)},
                                 headers={'D-Source': str(source.id), 'D-Destination': str(server.id)},
                                 verify=verify,
                                 timeout=timeout)
        except requests.exceptions.ReadTimeout as e:
            resp = None
            exc = e
        except requests.exceptions.ConnectionError as e:
            # unable to reach actual server through current gateway
            resp = None
            exc = e
        if resp is not None and resp.status_code == 200:
            cost = resp.json().get('hops', 0)
            elapsed = resp.elapsed
            tries = retries
    if exc:
        if isinstance(exc, requests.exceptions.ReadTimeout):
            logger.warning(f'Timeout reached while trying to access to {url}')
        elif isinstance(exc, requests.exceptions.ConnectionError):
            logger.debug(f'Unable to connect with {url}')
    return cost, elapsed


def _prepare_url(server: Server, view_or_url: str, view_data=None):
    if view_or_url.startswith('\\'):
        url = server.url() + 'view_or_url'
    else:
        view_data = view_data or {}
        url = server.url(view_or_url, **view_data)
    return url


def _prepare_headers(server, headers=None):
    headers = headers or {}
    headers.update({'D-Destination': str(server.id)})
    headers.update({'D-Source': str(Server.get_current().id)})
    return headers


def prepare_request(server, view_or_url, view_data, kwargs):
    params = kwargs.pop('params', {}) or {}

    url = _prepare_url(server, view_or_url, dict(**view_data, **params))

    kwargs['headers'] = _prepare_headers(server, kwargs.get('headers'))

    if 'auth' in kwargs and kwargs['auth'] is not None:
        kwargs['auth'](kwargs)

    if 'json' in kwargs and kwargs['json']:
        kwargs['json'] = pack_msg(kwargs['json'])

    return url


def request(method, server, view_or_url, view_data=None, session=None, **kwargs) -> Response:
    exception = None
    content = None
    status = None
    json_data = None
    headers = None

    if not session:
        _session = requests.session()
    else:
        _session = session

    raise_on_error = kwargs.pop('raise_on_error', False)

    try:
        url = prepare_request(server, view_or_url, view_data or {}, kwargs)
    except Exception as e:
        if raise_on_error:
            raise
        return Response(exception=e, server=server)

    func = getattr(_session, method.lower())

    if 'timeout' not in kwargs:
        kwargs['timeout'] = defaults.TIMEOUT_REQUEST

    kwargs['verify'] = current_app.config['SSL_VERIFY']

    try:
        resp: requests.Response = func(url, **kwargs)
    except (Timeout, ) as e:
        timeout = kwargs.get('timeout', None)
        if isinstance(timeout, tuple):
            timeout = timeout[0] + timeout[1]
        exception = TimeoutError(f"Socket timeout reached while trying to connect to {url} "
                                   f"for {timeout} seconds")
    except Exception as e:
        exception = e
    finally:
        if not session:
            _session.close()

    if exception is None:
        status = resp.status_code
        headers = resp.headers
        try:
            json_data = resp.json()
        except (ValueError,):
            content = resp.text

        if json_data:
            try:
                content = unpack_msg(json_data)
            except NotValidMessage:
                content = json_data
    else:
        if raise_on_error:
            raise exception

    return Response(msg=content, code=status, exception=exception, server=server, url=url, headers=headers)


def get(server: Server, view_or_url: str, view_data: Kwargs = None, session: requests.Session = None,
        params: Kwargs = None, **kwargs) -> Response:
    """Sends a GET request."""
    return request('get', server, view_or_url, view_data=view_data, session=session, params=params, **kwargs)


def options(server: Server, view_or_url: str, view_data: Kwargs = None, session: requests.Session = None,
            **kwargs) -> Response:
    r"""Sends an OPTIONS request."""

    return request('options', server, view_or_url, view_data=view_data, session=session, **kwargs)


def head(server: Server, view_or_url: str, view_data: Kwargs = None, session: requests.Session = None,
         **kwargs) -> Response:
    r"""Sends a HEAD request."""
    return request('head', server, view_or_url, view_data=view_data, session=session, **kwargs)


def post(server: Server, view_or_url: str, view_data: Kwargs = None, session: requests.Session = None, data=None,
         json=None, **kwargs) -> Response:
    r"""Sends a POST request."""
    return request('post', server, view_or_url, view_data=view_data, session=session, json=json or {}, **kwargs)


def put(server: Server, view_or_url: str, view_data: Kwargs = None, session: requests.Session = None, data=None,
        **kwargs) -> Response:
    r"""Sends a PUT request."""
    return request('put', server, view_or_url, view_data=view_data, session=session, **kwargs)


def patch(server: Server, view_or_url: str, view_data: Kwargs = None, session: requests.Session = None, data=None,
          **kwargs) -> Response:
    r"""Sends a PATCH request."""
    return request('patch', server, view_or_url, view_data=view_data, session=session, **kwargs)


def delete(server: Server, view_or_url: str, view_data: Kwargs = None, session: requests.Session = None,
           **kwargs) -> Response:
    r"""Sends a DELETE request."""
    return request('delete', server, view_or_url, view_data=view_data, session=session, **kwargs)


async def async_request(method, server, view_or_url, view_data=None, session=None, **kwargs) -> Response:
    exception = None
    content = None
    json_data = None
    status = None
    headers = None

    if not session:
        _session = aiohttp.ClientSession()
    else:
        _session = session

    raise_on_error = kwargs.pop('raise_on_error', False)

    try:
        url = prepare_request(server, view_or_url, view_data or {}, kwargs)
    except Exception as e:
        if raise_on_error:
            raise
        return Response(exception=e, server=server)

    kwargs['ssl'] = current_app.config['SSL_VERIFY']

    if 'timeout' in kwargs and isinstance(kwargs['timeout'], (int, float)):
        kwargs['timeout'] = aiohttp.ClientTimeout(total=kwargs['timeout'])

    func = getattr(_session, method)

    try:
        async with func(url, **kwargs) as resp:
            status = resp.status
            headers = resp.headers
            try:
                json_data = await resp.json()
            except (ContentTypeError, ValueError):
                content = await resp.text()
    except (concurrent.futures.TimeoutError, asyncio.TimeoutError) as e:
        exception = TimeoutError(f"Socket timeout reached while trying to connect to {url} "
                  f"for {kwargs.get('timeout').total or _session._timeout.total} seconds")
    except Exception as e:
        exception = e
    finally:
        if not session:
            await _session.close()

    if json_data:
        try:
            content = unpack_msg(json_data)
        except NotValidMessage:
            content = json_data
    elif exception:
        if raise_on_error:
            raise exception
    else:
        content = await resp.text()

    return Response(msg=content, code=status, exception=exception, server=server, url=url, headers=headers)


async def async_get(server: Server, view_or_url: str, view_data: Kwargs = None, session: aiohttp.ClientSession = None,
                    params: Kwargs = None, **kwargs) -> Response:
    """Sends a GET request."""
    return await async_request('get', server, view_or_url, view_data=view_data, session=session, params=params, **kwargs)


async def async_options(server: Server, view_or_url: str, view_data: Kwargs = None,
                        session: aiohttp.ClientSession = None,
                        **kwargs) -> Response:
    r"""Sends an OPTIONS request."""

    return await async_request('options', server, view_or_url, view_data=view_data, session=session, **kwargs)


async def async_head(server: Server, view_or_url: str, view_data: Kwargs = None, session: aiohttp.ClientSession = None,
                     **kwargs) -> Response:
    r"""Sends a HEAD request."""
    return await async_request('head', server, view_or_url, view_data=view_data, session=session, **kwargs)


async def async_post(server: Server, view_or_url: str, view_data: Kwargs = None, session: aiohttp.ClientSession = None,
                     json=None, **kwargs) -> Response:
    r"""Sends a POST request."""
    return await async_request('post', server, view_or_url, view_data=view_data, session=session, json=json,
                               **kwargs)


async def async_put(server: Server, view_or_url: str, view_data: Kwargs = None, session: aiohttp.ClientSession = None,
                    json=None, **kwargs) -> Response:
    r"""Sends a PUT request."""
    return await async_request('put', server, view_or_url, view_data=view_data, session=session, json=json, **kwargs)


async def async_patch(server: Server, view_or_url: str, view_data: Kwargs = None, session: aiohttp.ClientSession = None,
                      json=None, **kwargs) -> Response:
    r"""Sends a PATCH request."""
    return await async_request('patch', server, view_or_url, view_data=view_data, session=session, json=json, **kwargs)


async def async_delete(server: Server, view_or_url: str, view_data: Kwargs = None,
                       session: aiohttp.ClientSession = None,
                       **kwargs) -> Response:
    r"""Sends a DELETE request."""
    return await async_request('delete', server, view_or_url, view_data=view_data, session=session, **kwargs)



from flask import request
from flask_jwt_extended import jwt_required
from flask_restful import Resource

from dm.domain.entities.user import User
from dm.web import db
from dm.web.decorators import forward_or_dispatch, securizer, validate_schema, lock_catalog
from dm.web.helpers import filter_query
from dm.web.json_schemas import users_post, user_patch


class UserList(Resource):

    @forward_or_dispatch
    @jwt_required
    @securizer
    def get(self):
        query = filter_query(User, request.args, exclude=['_password'])
        return [user.to_json() for user in query.all()]

    @forward_or_dispatch
    @jwt_required
    @securizer
    @validate_schema(users_post)
    @lock_catalog
    def post(self):
        data = request.get_json()
        password = data.pop("password")
        u = User(**data)
        u.set_password(password)
        db.session.add(u)
        db.session.commit()
        return {'id': str(u.id)}, 201


class UserResource(Resource):
    @forward_or_dispatch
    @jwt_required
    @securizer
    def get(self, user_id):
        return User.query.get_or_404(user_id).to_json()

    @forward_or_dispatch
    @jwt_required
    @securizer
    @validate_schema(user_patch)
    @lock_catalog
    def patch(self, user_id):
        user = User.query.get_or_404(user_id)
        data = request.get_json()
        if 'email' in data and user.email != data.get('email'):
            user.email = data.get('email')
        if 'active' in data and user.active != data.get('active'):
            user.active = data.get('active')
        if user in db.session.dirty:
            db.session.commit()
            return {}, 204
        return {}, 202

    # @securizer
    # @jwt_required
    # @forward_or_dispatch
    # def delete(self, user_id):
    #     user = User.query.get_or_404(user_id)
    #     db.session.delete(user)
    #     db.session.commit()
    #     return {}, 204



from flask import url_for
from flask_jwt_extended import create_access_token

from dm.domain.entities import Orchestration, Step, ActionTemplate, ActionType
from dm.domain.entities.bootstrap import set_initial
from dm.network.auth import HTTPBearerAuth
from dm.web import create_app, db
from tests.helpers import TestCaseLockBypass


class Test(TestCaseLockBypass):
    def setUp(self):
        """Create and configure a new app instance for each test."""
        # create the app with common test config
        self.app = create_app('test')
        self.app_context = self.app.app_context()
        self.app_context.push()
        self.client = self.app.test_client()
        self.auth = HTTPBearerAuth(create_access_token('00000000-0000-0000-0000-000000000001'))
        self.o = Orchestration(name='name', version=1, description='desc')
        self.at = ActionTemplate(name='action_name', version=1, action_type=ActionType.SHELL, code='')

        db.session.add_all([self.o, self.at])
        set_initial()
        db.session.commit()

    def tearDown(self) -> None:
        db.session.remove()
        db.drop_all()
        self.app_context.pop()

    def test_step_resource_list(self):
        resp = self.client.get(url_for('api_1_0.steplist'), headers=self.auth.header)

        self.assertEqual([], resp.get_json())

        s = self.o.add_step(undo=True, action_template=self.at)
        db.session.add(s)
        db.session.commit()

        resp = self.client.get(url_for('api_1_0.steplist'), headers=self.auth.header)

        self.assertEqual([s.to_json()], resp.get_json())

    def test_step_resources(self):
        resp = self.client.post(url_for('api_1_0.steplist'),
                                json=dict(orchestration_id=str(self.o.id),
                                          undo=False,
                                          action_template_id=str(self.at.id)),
                                headers=self.auth.header)

        self.assertIn('step_id', resp.get_json())

        s_id = resp.get_json()['step_id']

        s = Step.query.get(s_id)

        resp = self.client.get(url_for('api_1_0.stepresource', step_id=s_id),
                               headers=self.auth.header)

        self.assertDictEqual(s.to_json(), resp.get_json())

        self.assertTrue(s.stop_on_error)
        self.assertTrue(s.stop_undo_on_error)
        resp = self.client.patch(url_for('api_1_0.stepresource', step_id=s_id),
                                 json={"stop_on_error": False},
                                 headers=self.auth.header)
        self.assertEqual(204, resp.status_code)
        db.session.refresh(s)
        self.assertFalse(s.stop_on_error)
        self.assertTrue(s.stop_undo_on_error)

        s2 = self.o.add_step(undo=True, action_template=self.at)
        s3 = self.o.add_step(undo=True, action_template=self.at)
        db.session.add_all([s2, s3])
        db.session.commit()

        resp = self.client.patch(url_for('api_1_0.stepresource', step_id=s_id),
                                 json={"child_step_ids": [str(s2.id)]},
                                 headers=self.auth.header)

        self.assertEqual(204, resp.status_code)

        resp = self.client.patch(url_for('api_1_0.stepresource', step_id=str(s3.id)),
                                 json={"parent_step_ids": [str(s.id)]},
                                 headers=self.auth.header)

        self.assertEqual(204, resp.status_code)

        db.session.refresh(s)
        self.assertEqual(2, len(s.children))
        self.assertIn(s2, s.children)
        self.assertIn(s3, s.children)

        resp = self.client.post(url_for('api_1_0.steplist'),
                                json=dict(orchestration_id=str(self.o.id),
                                          undo=False,
                                          action_template_id=str(self.at.id),
                                          parent_step_ids=[s_id],
                                          child_step_ids=[str(s2.id)]),
                                headers=self.auth.header)

        s4: Step = Step.query.get(resp.get_json()['step_id'])

        self.assertEqual(201, resp.status_code)
        self.assertListEqual([s2], s4.children)
        self.assertListEqual([s], s4.parents)

        resp = self.client.put(url_for('api_1_0.stepresource', step_id=str(s4.id)),
                               json=dict(undo=True,
                                         action_template_id=str(self.at.id),
                                         parent_step_ids=[str(s2.id)],
                                         child_step_ids=[str(s3.id)]),
                               headers=self.auth.header)

        db.session.refresh(s4)
        self.assertEqual(204, resp.status_code)
        self.assertListEqual([s2], s4.parents)
        self.assertListEqual([s3], s4.children)

        resp = self.client.delete(url_for('api_1_0.stepresource', step_id=s_id),
                               headers=self.auth.header)

        db.session.expire_all()
        self.assertEqual(204, resp.status_code)
        self.assertIsNone(Step.query.get(s_id))

    def test_step_resource_multiple_steps(self):
        resp = self.client.post(url_for('api_1_0.steplist'),
                                json=[dict(id=1, name="1", orchestration_id=str(self.o.id),
                                           undo=False,
                                           action_template_id=str(self.at.id)),
                                      dict(id=2, name="2", orchestration_id=str(self.o.id),
                                           undo=False,
                                           action_template_id=str(self.at.id),
                                           parent_step_ids=[1]),
                                      dict(id=3, name="3", orchestration_id=str(self.o.id),
                                           undo=False,
                                           action_template_id=str(self.at.id),
                                           parent_step_ids=[2]),
                                      dict(id=4, name="4", orchestration_id=str(self.o.id),
                                           undo=False,
                                           action_template_id=str(self.at.id),
                                           parent_step_ids=[1],
                                           child_step_ids=[3])
                                      ],
                                headers=self.auth.header)

        self.assertEqual(201, resp.status_code)
        db.session.refresh(self.o)
        s1, = self.o.root
        self.assertEqual(0, len(s1.parents))
        self.assertEqual(2, len(s1.children))

        s2, s4 = s1.children

        self.assertEqual(1, len(s2.parents))
        self.assertEqual(1, len(s2.children))

        self.assertEqual(1, len(s4.parents))
        self.assertEqual(1, len(s4.children))



import threading
from unittest import TestCase

from dm import defaults
from dm.domain.entities import Server, Dimension, Software, Transfer, TransferStatus
from dm.utils.helpers import get_now
from dm.web import create_app, db


class TestTransfer(TestCase):
    def setUp(self):
        """Create and configure a new app instance for each test."""
        # create the app with common test config
        self.app = create_app('test')
        self.app_context = self.app.app_context()
        self.app_context.push()
        db.create_all()
        Server.set_initial()
        d = Dimension(name='test', current=True)
        db.session.add(d)
        db.session.commit()
        self.client = self.app.test_client(use_cookies=True)

    def tearDown(self) -> None:
        db.session.remove()
        db.drop_all()
        self.app_context.pop()

    def test_wait_transfer(self):
        s = Software(name='test', version='1', filename='file')
        t = Transfer(software=s, dest_path='', num_chunks=0)

        db.session.add_all([s, t])
        db.session.commit()

        status = t.wait_transfer(timeout=0.1, refresh_interval=0.01)
        self.assertEqual(TransferStatus.WAITING_CHUNKS, status)

        def update_transfer(app, transfer_id):
            with app.app_context():
                tr = Transfer.query.get(transfer_id)
                tr.status = TransferStatus.COMPLETED
                db.session.commit()

        th = threading.Timer(0.05, update_transfer, (self.app, t.id))
        th.start()
        status = t.wait_transfer(timeout=0.1, refresh_interval=0.01)
        self.assertEqual(TransferStatus.COMPLETED, status)

    def test_to_json(self):
        d = get_now()
        s = Software(id='aaaaaaaa-1234-5678-1234-56781234aaa1', name='test', version='1', filename='file')
        t = Transfer(id='aaaaaaaa-1234-5678-1234-56781234aaa2', software=s, dest_path='/folder', num_chunks=0, created_on=d)

        self.assertDictEqual(
            dict(id='aaaaaaaa-1234-5678-1234-56781234aaa2', software_id='aaaaaaaa-1234-5678-1234-56781234aaa1',
                 dest_path='/folder', num_chunks=0, file='/folder/file',
                 status='WAITING_CHUNKS', created_on=d.strftime(defaults.DATETIME_FORMAT)),
            t.to_json())

        t = Transfer(id='aaaaaaaa-1234-5678-1234-56781234aaa2', software='filename', size=10, checksum='abc12',
                     dest_path='/folder', num_chunks=0, created_on=d)

        self.assertDictEqual(
            dict(id='aaaaaaaa-1234-5678-1234-56781234aaa2', size=10, checksum='abc12',
                 dest_path='/folder', num_chunks=0, file='/folder/filename',
                 status='WAITING_CHUNKS', created_on=d.strftime(defaults.DATETIME_FORMAT)),
            t.to_json())



import configparser
import datetime as dt
import hashlib
import inspect
import itertools
import logging
import os
import platform
import re
import sys
import typing as t
import uuid
from collections import Iterable, ChainMap

import netifaces
import six
import yaml
from cryptography.fernet import Fernet
from flask import current_app
from tzlocal import get_localzone

import dm.defaults as d
from dm import defaults


class AttributeDict(dict):
    __getattr__ = dict.__getitem__
    __setattr__ = dict.__setitem__


def is_string_types(var):
    return isinstance(var, six.string_types)


def is_iterable(var):
    return isinstance(var, Iterable)


def is_iterable_not_string(var):
    return is_iterable(var) and not is_string_types(var)


def is_valid_uuid(var):
    try:
        uuid.UUID(var)
    except ValueError:
        return False
    else:
        return True


def convert(d):
    for k, v in d.items():
        if isinstance(v, dict):
            d[k] = convert(v)
    return AttributeDict(d)


# function to mock datetime.now
def get_now() -> dt.datetime:
    return dt.datetime.now(get_localzone())


def str_to_key(id_: str):
    key = []
    for x in id_.split('.'):
        try:
            x = int(x)
        except ValueError:
            pass
        finally:
            key.append(x)
    return tuple(key) if len(key) > 1 else key[0]


def key_to_str(key):
    if isinstance(key, t.Sequence):
        id_ = '.'.join([str(i) for i in key])
    else:
        id_ = str(key)
    return id_


def generate_url(destination, uri, protocol='https'):
    from dm.domain.entities import Server
    try:
        forwarder = destination.mesh_best_route[0]
    except IndexError:
        forwarder = destination
    else:
        forwarder = Server.query.get(forwarder)

    return f"{protocol}://{forwarder.name}:{forwarder.port}{uri}"


def generate_symmetric_key():
    return Fernet.generate_key()


def encrypt_symmetric(data, key):
    cipher_suite = Fernet(key)
    return cipher_suite.encrypt(data)


def decrypt_symmetric(data, key):
    cipher_suite = Fernet(key)
    return cipher_suite.decrypt(data)


def encrypt(data: bytes, symmetric_key: bytes = None) -> \
        t.Tuple[bytes, t.Optional[bytes]]:
    """

    Parameters
    ----------
    data:
        data to encrypt.
    symmetric_key:
        symmetric key used for encrypting data. If set, cipher_key must be None

    Returns
    -------

    """
    new_symmetric_key = None
    if not symmetric_key:
        symmetric_key = new_symmetric_key = generate_symmetric_key()
    cipher_data = encrypt_symmetric(data, symmetric_key)
    return cipher_data, new_symmetric_key


def decrypt(cipher_text: bytes, symmetric_key: bytes) -> bytes:
    """

    Parameters
    ----------
    cipher_text:
        text to decrypt
    cipher_key:
        symmetric_key encrypted. If specified symmetric_key must be None (default)
    symmetric_key:
        symmetric_key. If specified cipher_token must be None (default)
    key:
        key used for decryption of cipher_key

    Returns
    -------
    decrypted data
    """
    cipher_suite = Fernet(symmetric_key)
    dumped_data = cipher_suite.decrypt(cipher_text)
    return dumped_data


def get_logger(self=None):
    if self:
        name = '.'.join([
            self.__module__,
            self.__name__
        ])
    else:
        name = None
    try:
        current_app.logger
    except RuntimeError:
        logger = logging.getLogger(name)
    else:
        logger = current_app.logger
    return logger


def get_entities() -> t.List[t.Tuple['str', t.Any]]:
    from dm.web import db
    entities = []
    for name, cls in inspect.getmembers(sys.modules['dm.domain.entities'],
                                        lambda x: (inspect.isclass(x) and issubclass(x, db.Model))):
        entities.append((name, cls))

    return entities


def get_distributed_entities() -> t.List[t.Tuple['str', t.Any]]:
    from dm.domain.entities.base import DistributedEntityMixin
    entities = []
    for name, cls in inspect.getmembers(sys.modules['dm.domain.entities'],
                                        lambda x: (inspect.isclass(x) and issubclass(x, DistributedEntityMixin))):
        entities.append((name, cls))

    return sorted(entities, key=lambda x: x[1].order or 99999)


def md5(fname):
    hash_md5 = hashlib.md5()
    with open(fname, "rb") as f:
        for chunk in iter(lambda: f.read(4096), b""):
            hash_md5.update(chunk)
    return hash_md5.hexdigest()


def load_config_yaml():
    filename = os.path.join(d.HOME, d.CONFIG_FILE)
    with open(filename) as fd:
        config = yaml.load(fd, Loader=yaml.FullLoader)
    # if 'host' in config and config['host'].strip() == '0.0.0.0':
    #     config['host'] = '127.0.0.1'

    return config


def save_config_yaml(config):
    filename = os.path.join(d.HOME, d.CONFIG_FILE)
    with open(filename, 'w') as fd:
        yaml.dump(config, fd)


def update_config_yaml(param, value):
    filename = os.path.join(d.HOME, d.CONFIG_FILE)
    yaml_config = load_config_yaml()
    attributes = param.split('.')
    selected = yaml_config
    i = 0
    while i < (len(attributes) - 1):
        selected = selected[attributes[i]]
        i += 1
    selected.update({attributes[-1]: value})
    save_config_yaml(yaml_config)


def load_config_wsgi():
    filename = os.path.join(d.HOME, d.WSGI_FILE)
    config = configparser.ConfigParser()
    # if not os.path.exists(filename):
    #     file = os.path.join(os.getcwd(), filename)
    # else:
    r = config.read(filename)
    protocol = None
    if len(r) == 0:
        raise FileNotFoundError(f"unable to find file '{filename}'")
    if 'uwsgi' not in config:
        raise ValueError("Section 'uwsgi' not found in file")
    if 'https' in config['uwsgi']:
        raise NotImplemented('https not supported')
    else:
        if 'http' in config['uwsgi']:
            protocol = 'http'
            ip, port = config['uwsgi']['http'].split(':')

    if ip is None or ip == '*' or ip == '0.0.0.0':
        ip = '127.0.0.1'
    port = int(port) if port else defaults.PORT
    return {'dm': {'protocol': protocol, 'host': ip, 'port': port, 'venv': config.get('venv', None)}}


def collect_initial_config():
    config = ChainMap(load_config_yaml())
    # if platform.system() == 'Windows':
    #     p = find_process_by_name('flask')
    #     if not p:
    #         p = find_python_file_executed('dimensigon.py')
    #     args = p.cmdline()
    #     host_op = '-h' if '-h' in args else '--host' if '--host' in args else None
    #     if host_op:
    #         ip = args[args.index(host_op) + 1]
    #     else:
    #         ip = '127.0.0.1'
    #     port_op = '-p' if '-p' in args else '--port' if '--port' in args else None
    #     if port_op:
    #         port = args[args.index(port_op) + 1]
    #     else:
    #         if 'FLASK_RUN_PORT' in p.environ():
    #             port = p.environ()['FLASK_RUN_PORT']
    #         else:
    #             port = 5000
    #     protocol = 'https' if '--cert' in args else 'http'
    #
    #     config = {'protocol': protocol, 'ip': ip, 'port': port, 'venv': p.environ().get('VIRTUAL_ENV', None)}
    if platform.system() == 'Linux':
        config = config.new_child(load_config_wsgi())

    return config


def get_filename_from_cd(cd):
    """
    Get filename from content-disposition
    """
    if not cd:
        return None
    fname = re.findall('filename=(.+)', cd)
    if len(fname) == 0:
        return None
    return fname[0]


def generate_dimension(name):
    from cryptography.hazmat.backends import default_backend
    from cryptography.hazmat.primitives import serialization
    from cryptography.hazmat.primitives.asymmetric import rsa
    from dm.domain.entities import Dimension

    private_key = rsa.generate_private_key(
        public_exponent=65537,
        key_size=4096,
        backend=default_backend()
    )
    priv_pem = private_key.private_bytes(encoding=serialization.Encoding.PEM,
                                         format=serialization.PrivateFormat.TraditionalOpenSSL,
                                         encryption_algorithm=serialization.NoEncryption())
    pub_pem = private_key.public_key().public_bytes(encoding=serialization.Encoding.PEM,
                                                    format=serialization.PublicFormat.PKCS1)

    return Dimension(name=name, private=priv_pem, public=pub_pem)


def remove_prefix(s, prefix):
    return s[len(prefix):] if s.startswith(prefix) else s


def get_ips_listening_for() -> t.List[t.Tuple[str, int]]:
    from gunicorn_conf import bind
    gates = []
    for b in bind:
        dns_or_ip, port = b.split(':')
        if dns_or_ip == '0.0.0.0':
            ips = list(itertools.chain(
                *[[ip['addr'] for ip in netifaces.ifaddresses(iface).get(netifaces.AF_INET, [])] for iface in
                  netifaces.interfaces()]))
            gates.extend([(ip, port) for ip in ips])
        else:
            gates.append((dns_or_ip, port))
    return gates


def clean_string(incoming_string):
    replace_char = '_'
    newstring = incoming_string
    newstring = newstring.replace("!", replace_char)
    newstring = newstring.replace("@", replace_char)
    newstring = newstring.replace("#", replace_char)
    newstring = newstring.replace("$", replace_char)
    newstring = newstring.replace("%", replace_char)
    newstring = newstring.replace("^", replace_char)
    newstring = newstring.replace("&", replace_char)
    newstring = newstring.replace("*", replace_char)
    newstring = newstring.replace("(", replace_char)
    newstring = newstring.replace(")", replace_char)
    newstring = newstring.replace("+", replace_char)
    newstring = newstring.replace("=", replace_char)
    newstring = newstring.replace("?", replace_char)
    newstring = newstring.replace("\'", replace_char)
    newstring = newstring.replace("\"", replace_char)
    newstring = newstring.replace("{", replace_char)
    newstring = newstring.replace("}", replace_char)
    newstring = newstring.replace("[", replace_char)
    newstring = newstring.replace("]", replace_char)
    newstring = newstring.replace("<", replace_char)
    newstring = newstring.replace(">", replace_char)
    newstring = newstring.replace("~", replace_char)
    newstring = newstring.replace("`", replace_char)
    newstring = newstring.replace(":", replace_char)
    newstring = newstring.replace(";", replace_char)
    newstring = newstring.replace("|", replace_char)
    newstring = newstring.replace("\\", replace_char)
    newstring = newstring.replace("/", replace_char)
    newstring = newstring.replace("-", replace_char)
    return newstring



import json
from unittest import TestCase

import responses
from aioresponses import CallbackResult, aioresponses
from flask_jwt_extended import create_access_token
from requests import PreparedRequest
from requests.exceptions import ConnectTimeout

from dm.domain.entities import Server, Dimension
from dm.network.auth import HTTPBearerAuth
from dm.utils import asyncio
from dm.utils.asyncio import run
from dm.utils.helpers import generate_dimension
from dm.web import create_app, db
from dm.web.decorators import securizer
from dm.web.network import get, post, async_get, async_post, Response, pack_msg, unpack_msg
from tests.helpers import generate_dimension_json_data

healthcheck_view = 'root.healthcheck'


class TestNetwork(TestCase):

    @classmethod
    def setUpClass(cls) -> None:
        cls.json_dim = generate_dimension_json_data()

    def setUp(self):
        """Create and configure a new app instance for each test."""
        # create the app with common test config
        self.app = create_app('test')

        @self.app.route('/', methods=['GET', 'POST'])
        def home():
            return {'msg': 'default response'}

        @self.app.route('/securized', methods=['GET', 'POST'])
        @securizer
        def securized():
            return {'msg': 'default response'}

        self.app_context = self.app.app_context()
        self.app_context.push()
        self.client = self.app.test_client()
        db.init_app(self.app)
        db.create_all()
        self.server = Server('me', port=5000, me=True)
        db.session.add(self.server)
        db.session.add(Dimension.from_json(self.json_dim))
        db.session.commit()

        self.token = create_access_token('test')
        self.auth = HTTPBearerAuth(self.token)
        self.url = 'http://me:5000/'

    def tearDown(self) -> None:
        db.session.remove()
        db.drop_all()
        self.app_context.pop()

    @aioresponses()
    @responses.activate
    def test_get_headers_url_response(self, m):
        msg = {'data': 'content'}
        status = 200

        def callback(request, **kwargs):
            if isinstance(request, PreparedRequest):
                self.assertEqual(str(self.server.id), request.headers.get('D-Destination'))
                self.assertEqual(str(self.server.id), request.headers.get('D-Source'))
                self.assertEqual("True", request.headers.get('D-Securizer'))
                self.assertEqual(f"Bearer {self.token}", request.headers.get('Authorization'))

                return status, {}, json.dumps(msg)
            else:
                self.assertEqual(str(self.server.id), kwargs['headers'].get('D-Destination'))
                self.assertEqual(str(self.server.id), kwargs['headers'].get('D-Source'))
                self.assertEqual("True", kwargs['headers'].get('D-Securizer'))
                self.assertEqual(f"Bearer {self.token}", kwargs['headers'].get('Authorization'))

                return CallbackResult(status=status, payload=msg)

        responses.add_callback(responses.GET, self.url, callback=callback)
        m.get(self.url, callback=callback)

        resp = get(self.server, 'home', auth=self.auth, headers={'D-Securizer': "True"})

        self.assertEqual(
            Response(msg=msg, code=status, exception=None, server=self.server, url=self.url),
            resp)

        resp = run(async_get(self.server, 'home', auth=self.auth, headers={'D-Securizer': "True"}))

        self.assertEqual(
            Response(msg=msg, code=status, exception=None, server=self.server, url=self.url),
            resp)

    @aioresponses()
    @responses.activate
    def test_get_error_json(self, m):
        msg = {'error': 'this is an error message'}
        status = 500
        responses.add(responses.GET, self.url, json=msg, status=status)
        m.get(self.url, payload=msg, status=status)

        resp = get(self.server, 'home')

        self.assertEqual(status, resp.code)
        self.assertDictEqual(msg, resp.msg)

        resp = run(async_get(self.server, 'home'))

        self.assertEqual(status, resp.code)
        self.assertDictEqual(msg, resp.msg)

    @aioresponses()
    @responses.activate
    def test_get_internal_error_server(self, m):
        msg = '<html>Iternal error server</html>'
        status = 500

        responses.add(responses.GET, self.url, status=status, body=msg)
        m.get(self.url, status=status, body=msg)

        resp = get(Server.get_current(), 'home')

        self.assertEqual(status, resp.code)
        self.assertEqual(msg, resp.msg)

        resp = run(async_get(Server.get_current(), 'home'))

        self.assertEqual(status, resp.code)
        self.assertEqual(msg, resp.msg)

    @aioresponses()
    @responses.activate
    def test_post(self, m):
        status = 200
        msg = {'new': 'data'}
        data = {'data': 'some data'}

        def callback(request, **kwargs):
            if isinstance(request, PreparedRequest):
                self.assertDictEqual(data, json.loads(request.body))
                return status, {}, json.dumps(msg)
            else:
                self.assertDictEqual(kwargs['json'], data)
                return CallbackResult(status=status, payload=msg)

        responses.add_callback(responses.POST, self.url, callback=callback)
        m.post(self.url, callback=callback)

        resp = post(self.server, 'home', json=data)

        self.assertEqual(status, resp.code)
        self.assertDictEqual(msg, resp.msg)

        resp = run(async_post(self.server, 'home', json=data))

        self.assertEqual(status, resp.code)
        self.assertDictEqual(msg, resp.msg)

    @aioresponses()
    @responses.activate
    def test_post_no_content_in_response(self, m):
        msg = ''
        status = 204
        responses.add(responses.POST, self.url, status=204)
        m.post(self.url, status=204)

        data, status = post(self.server, 'home')

        self.assertEqual(status, status)
        self.assertEqual(msg, data)

        data, status = run(async_post(self.server, 'home'))

        self.assertEqual(status, status)
        self.assertEqual(msg, data)

    @aioresponses()
    @responses.activate
    def test_connection_error(self, m):
        responses.add(responses.GET, self.url, body=ConnectionError())
        m.get(self.url, exception=ConnectionError())

        resp = get(self.server, 'home')

        self.assertIsNone(resp.code)
        self.assertIsNone(resp.msg)
        self.assertIsInstance(resp.exception, ConnectionError)

        resp = run(async_get(self.server, 'home'))

        self.assertIsNone(resp.code)
        self.assertIsNone(resp.msg)
        self.assertIsInstance(resp.exception, ConnectionError)

    @aioresponses()
    @responses.activate
    def test_timeout(self, m):
        responses.add(responses.GET, self.url, body=ConnectTimeout())
        m.get(self.url, exception=asyncio.TimeoutError())

        resp = get(self.server, 'home', timeout=0.01)

        self.assertIsNone(resp.code)
        self.assertIsNone(resp.msg)
        self.assertIsInstance(resp.exception, TimeoutError)
        self.assertEqual(f"Socket timeout reached while trying to connect to http://me:5000/ "
                         f"for 0.01 seconds", str(resp.exception))

        resp = run(async_get(self.server, 'home', timeout=0.01))

        self.assertIsNone(resp.code)
        self.assertIsNone(resp.msg)
        self.assertIsInstance(resp.exception, TimeoutError)
        self.assertEqual(f"Socket timeout reached while trying to connect to http://me:5000/ "
                         f"for 0.01 seconds", str(resp.exception))

    @aioresponses()
    @responses.activate
    def test_raise_on_error(self, m):
        responses.add(responses.GET, self.url, body=ConnectTimeout())
        m.get(self.url, exception=asyncio.TimeoutError())

        with self.assertRaises(TimeoutError) as e:
            resp = get(self.server, 'home', timeout=1, raise_on_error=True)

        self.assertEqual(f"Socket timeout reached while trying to connect to {Server.get_current().url('root.home')} "
                         f"for 1 seconds", str(e.exception))

        with self.assertRaises(TimeoutError) as e:
            resp = run(
                async_get(self.server, 'home', timeout=1, raise_on_error=True))

        self.assertEqual(f"Socket timeout reached while trying to connect to {Server.get_current().url('root.home')} "
                         f"for 1 seconds", str(e.exception))

    @aioresponses()
    @responses.activate
    def test_raise_on_error_url(self, m):
        responses.add(responses.GET, self.url, body=ConnectTimeout())
        m.get(self.url, exception=asyncio.TimeoutError())

        self.server.gates = []

        with self.assertRaises(RuntimeError) as e:
            resp = get(self.server, 'home', raise_on_error=True)

        resp = get(self.server, 'home')

        self.assertIsInstance(resp.exception, RuntimeError)

        with self.assertRaises(RuntimeError) as e:
            resp = run(
                async_get(self.server, 'home', raise_on_error=True))

        resp = run(
            async_get(self.server, 'home'))

        self.assertIsInstance(resp.exception, RuntimeError)


class TestNetworkSecurizer(TestCase):

    def setUp(self):
        """Create and configure a new app instance for each test."""
        # create the app with common test config
        self.app = create_app('test')
        self.app.config['SECURIZER'] = True

        @self.app.route('/', methods=['GET', 'POST'])
        @securizer
        def home():
            return {'msg': 'default response'}

        self.app_context = self.app.app_context()
        self.app_context.push()
        self.client = self.app.test_client()
        db.init_app(self.app)
        db.create_all()
        self.server = Server('me', port=5000, me=True)
        db.session.add(self.server)

        self.dim = generate_dimension('dimension')
        self.dim.current = True
        db.session.add(self.dim)

        self.token = create_access_token('test')
        self.auth = HTTPBearerAuth(self.token)
        self.url = 'http://me:5000/'

    def tearDown(self) -> None:
        db.session.remove()
        db.drop_all()
        self.app_context.pop()

    @aioresponses()
    @responses.activate
    def test_get_headers_url_response(self, m):
        msg = {'data': 'content'}
        status = 200

        def callback(request, **kwargs):
            if isinstance(request, PreparedRequest):
                self.assertEqual(str(self.server.id), request.headers.get('D-Destination'))
                self.assertEqual(str(self.server.id), request.headers.get('D-Source'))
                self.assertEqual("True", request.headers.get('D-Securizer'))
                self.assertEqual(f"Bearer {self.token}", request.headers.get('Authorization'))

                return status, {}, json.dumps(pack_msg(msg))
            else:
                self.assertEqual(str(self.server.id), kwargs['headers'].get('D-Destination'))
                self.assertEqual(str(self.server.id), kwargs['headers'].get('D-Source'))
                self.assertEqual("True", kwargs['headers'].get('D-Securizer'))
                self.assertEqual(f"Bearer {self.token}", kwargs['headers'].get('Authorization'))

                return CallbackResult(status=status, payload=pack_msg(msg))

        responses.add_callback(responses.GET, self.url, callback=callback)
        m.get(self.url, callback=callback)

        resp = get(self.server, 'home', auth=self.auth, headers={'D-Securizer': "True"})

        self.assertEqual(
            Response(msg=msg, code=status, exception=None, server=self.server, url=self.url),
            resp)

        resp = run(async_get(self.server, 'home', auth=self.auth, headers={'D-Securizer': "True"}))

        self.assertEqual(
            Response(msg=msg, code=status, exception=None, server=self.server, url=self.url),
            resp)

    @aioresponses()
    @responses.activate
    def test_get_error_json(self, m):
        msg = {'error': 'this is an error message'}
        status = 500
        responses.add(responses.GET, self.url, json=pack_msg(msg), status=status)
        m.get(self.url, payload=pack_msg(msg), status=status)

        resp = get(self.server, 'home')

        self.assertEqual(status, resp.code)
        self.assertDictEqual(msg, resp.msg)

        resp = run(async_get(self.server, 'home'))

        self.assertEqual(status, resp.code)
        self.assertDictEqual(msg, resp.msg)

    @aioresponses()
    @responses.activate
    def test_get_internal_error_server(self, m):
        msg = '<html>Iternal error server</html>'
        status = 500

        responses.add(responses.GET, self.url, status=status, body=msg)
        m.get(self.url, status=status, body=msg)

        resp = get(Server.get_current(), 'home')

        self.assertEqual(status, resp.code)
        self.assertEqual(msg, resp.msg)

        resp = run(async_get(Server.get_current(), 'home'))

        self.assertEqual(status, resp.code)
        self.assertEqual(msg, resp.msg)

    @aioresponses()
    @responses.activate
    def test_post(self, m):
        status = 200
        msg = {'new': 'data'}
        data = {'data': 'some data'}

        def callback(request, **kwargs):
            if isinstance(request, PreparedRequest):
                self.assertDictEqual(data, unpack_msg(json.loads(request.body)))
                return status, {}, json.dumps(pack_msg(msg))
            else:
                self.assertDictEqual(data, unpack_msg(kwargs['json']))
                return CallbackResult(status=status, payload=pack_msg(msg))

        responses.add_callback(responses.POST, self.url, callback=callback)
        m.post(self.url, callback=callback)

        resp = post(self.server, 'home', json=data)

        self.assertEqual(status, resp.code)
        self.assertDictEqual(msg, resp.msg)

        resp = run(async_post(self.server, 'home', json=data))

        self.assertEqual(status, resp.code)
        self.assertDictEqual(msg, resp.msg)

    @aioresponses()
    @responses.activate
    def test_post_no_content_in_response(self, m):
        msg = ''
        status = 204
        responses.add(responses.POST, self.url, status=204)
        m.post(self.url, status=204)

        data, status = post(self.server, 'home')

        self.assertEqual(status, status)
        self.assertEqual(msg, data)

        data, status = run(async_post(self.server, 'home'))

        self.assertEqual(status, status)
        self.assertEqual(msg, data)



import argparse
import copy
import os
import shlex
from pprint import pprint

from prompt_toolkit import prompt
from prompt_toolkit.auto_suggest import AutoSuggestFromHistory
from prompt_toolkit.completion import merge_completers
from prompt_toolkit.history import InMemoryHistory
from prompt_toolkit.lexers import PygmentsLexer
from pygments.lexers.python import PythonLexer

import dm.dshell.network as ntwrk
from dm.domain.entities import ActionType
from dm.dshell.argparse_raise import ArgumentParserRaise
from dm.dshell.completer import granule_completer, server_name_completer, action_completer
from dm.dshell.helpers import get_history, exit_dshell
from dm.dshell.prompts.action_template import code_lexer
from dm.dshell.prompts.utils import prompt_parameter
from dm.dshell.validators import ChoiceValidator, IntValidator, JSONValidator, BoolValidator, ListValidator

form = {
    "undo": dict(history=InMemoryHistory()),
    "action_template_id": dict(history=InMemoryHistory(), completer=action_completer),
    "stop_on_error": dict(validator=BoolValidator, history=InMemoryHistory()),
    "stop_undo_on_error": dict(validator=BoolValidator, history=InMemoryHistory()),
    "undo_on_error": dict(validator=BoolValidator, history=InMemoryHistory()),
    "action_type": dict(validator=ChoiceValidator([at.name for at in ActionType if at.name != 'NATIVE']),
                        history=InMemoryHistory()),
    "code": dict(multiline=True, lexer=code_lexer, history=InMemoryHistory()),
    "expected_stdout": dict(multiline=True, history=InMemoryHistory()),
    "expected_stderr": dict(multiline=True, history=InMemoryHistory()),
    "expected_rc": dict(validator=IntValidator(), history=InMemoryHistory()),
    "parameters": dict(multiline=True, lexer=PygmentsLexer(PythonLexer), validator=JSONValidator(),
                       history=InMemoryHistory()),
    "system_kwargs": dict(multiline=True, lexer=PygmentsLexer(PythonLexer), validator=JSONValidator(),
                          history=InMemoryHistory()),
    "pre_process": dict(multiline=True, lexer=PygmentsLexer(PythonLexer), history=InMemoryHistory()),
    "post_process": dict(multiline=True, lexer=PygmentsLexer(PythonLexer), history=InMemoryHistory()),
    "target": dict(completer=merge_completers([granule_completer, server_name_completer]), history=InMemoryHistory()),
    "parent_step_ids": dict(validator=ListValidator())
}

parser = ArgumentParserRaise(allow_abbrev=False, prog='')
subparser = parser.add_subparsers(dest='cmd')
preview_parser = subparser.add_parser('preview')
subpreview_parser = preview_parser.add_subparsers(dest='subcmd')
subpreview_parser.add_parser('action')
set_parser = subparser.add_parser('set')
set_parser.add_argument('parameter', help="Available: " + ', '.join(form.keys()))
save_parser = subparser.add_parser('save')
exit_parser = subparser.add_parser('exit')

history = None

entity_name = os.path.basename(__file__).split('.py')[0]


def subprompt(entity, changed=False, ask_all=False, parent_prompt=None):
    global history
    if history is None:
        history = get_history(entity_name, InMemoryHistory())

    entity = copy.deepcopy(entity)
    if ask_all:
        for k, v in form.items():
            if k not in entity:
                try:
                    if prompt_parameter(k, entity, form, f"{parent_prompt}:{entity['id']}"):
                        changed = True
                except KeyboardInterrupt:
                    entity[k] = None
                except EOFError:
                    exit_dshell()

    while True:
        try:
            text = prompt(f"{parent_prompt}:{entity['id']}> ", history=history,
                          auto_suggest=AutoSuggestFromHistory())
        except KeyboardInterrupt:
            continue
        except EOFError:
            exit_dshell(rc=1)
        try:
            namespace = parser.parse_args(shlex.split(text))
        except (ValueError, argparse.ArgumentError) as e:
            print(e)
            continue
        except SystemExit:
            continue

        if namespace.cmd == 'preview':
            if namespace.subcmd == 'action':
                if entity.get('action_template_id', None):
                    resp = ntwrk.get('api_1_0.actiontemplateresource', view_data={'action_template_id': entity['action_template_id']})
                    if resp.ok:
                        pprint(resp.msg)
                    else:
                        if resp.code is not None:
                            pprint(resp.msg)
                        else:
                            print(str(resp.exception))
                else:
                    print('no action_template_id set')
            elif namespace.subcmd is None:
                pprint(entity, indent=4)
        elif namespace.cmd == 'set':
            if namespace.parameter not in form.keys():
                print("Not a valid parameter. Available: " + ', '.join(form.keys()))
            else:
                try:
                    if prompt_parameter(namespace.parameter, entity, form, f"{parent_prompt}:{entity['id']}"):
                        changed = True
                except KeyboardInterrupt:
                    continue  # Control-C pressed. Try again.
                except EOFError:
                    exit_dshell(rc=1)
        elif namespace.cmd == 'save':
            if changed:
                return entity
            else:
                return None
        elif namespace.cmd == 'exit':
            if changed:
                text = ''
                while text.lower().strip() not in ('y', 'n'):
                    text = prompt('If you exit you will lose the changes. Do you want to continue? (y/n): ')
                if text.lower().strip() == 'y':
                    break
            else:
                break



from config import Config

data1 = {
    'ActionTemplateRepo': [
        dict(id='aaaaaaaa-1234-5678-1234-56781234aaa1', name='mkdir', version=1, action_type='SHELL',
             code='mkdir {dir}', parameters={}, expected_output=None, expected_rc=None,
             system_kwargs={}, data_mark='20190101000530100000'),
        dict(id='aaaaaaaa-1234-5678-1234-56781234aaa2', name='rmdir', version=1, action_type='SHELL',
             code='rmdir {dir}', parameters={}, expected_output=None, expected_rc=None,
             system_kwargs={}, data_mark='20190101000530100000')
    ],
    'StepRepo': [
        dict(id='eeeeeeee-1234-5678-1234-56781234eee1', undo=False, stop_on_error=True,
             action_template='aaaaaaaa-1234-5678-1234-56781234aaa1',
             step_expected_output=None, step_expected_rc=0, step_parameters={'dir': 'folder'}, step_system_kwargs=None,
             data_mark='20190101000530100000'),
        dict(id='eeeeeeee-1234-5678-1234-56781234eee2', undo=True, stop_on_error=True,
             action_template='aaaaaaaa-1234-5678-1234-56781234aaa2',
             step_expected_output=None, step_expected_rc=0, step_parameters={'dir': 'folder'},
             step_system_kwargs=None, data_mark='20190101000530100000')
    ],
    'OrchestrationRepo': [
        dict(id='cccccccc-1234-5678-1234-56781234ccc1', name='create folder', version=1,
             description='Creates a folder on the specified location',
             steps=['eeeeeeee-1234-5678-1234-56781234eee1',
                    'eeeeeeee-1234-5678-1234-56781234eee2'],
             dependencies={'eeeeeeee-1234-5678-1234-56781234eee1': ['eeeeeeee-1234-5678-1234-56781234eee2'],
                           'eeeeeeee-1234-5678-1234-56781234eee2': []},
             data_mark='20190101000530100000')
    ],
    'ServerRepo': [
        dict(id='bbbbbbbb-1234-5678-1234-56781234bbb1', name='server1.localdomain', ip='127.0.0.1', port=80, birth=None,
             keep_alive=None, available=True, granules=[], route=[], alt_route=[], data_mark='20190101000530100000'),
        dict(id='bbbbbbbb-1234-5678-1234-56781234bbb2', name='server2.localdomain', ip='127.0.0.1', port=81, birth=None,
             keep_alive=None, available=True, granules=[], route=[], alt_route=[], data_mark='20190101000530100000'),
        dict(id='bbbbbbbb-1234-5678-1234-56781234bbb3', name='server3.localdomain', ip='127.0.0.1', port=80, birth=None,
             keep_alive=None, available=True, granules=[], route=['bbbbbbbb-1234-5678-1234-56781234bbb2'],
             alt_route=[], data_mark='20190101000530100000'),
    ],
    'CatalogRepo': [
        dict(entity='ActionTemplate', data_mark='20190101000530100000'),
        dict(entity='Step', data_mark='20190101000530100000'),
        dict(entity='Orchestration', data_mark='20190101000530100000'),
        dict(entity='Server', data_mark='20190101000530100000'),
    ]
}

data2 = {'ActionTemplateRepo': [
    dict(id='aaaaaaaa-1234-5678-1234-56781234aaa1', name='mkdir', version=1, action_type='SHELL',
         code='mkdir {dir}', parameters={}, expected_output=None, expected_rc=None,
         system_kwargs={}, data_mark='20190101000532100000'),
    dict(id='aaaaaaaa-1234-5678-1234-56781234aaa2', name='rmdir', version=1, action_type='SHELL',
         code='rmdir {dir}', parameters={}, expected_output=None, expected_rc=None,
         system_kwargs={}, data_mark='20190101000530100000'),
    dict(id='aaaaaaaa-1234-5678-1234-56781234aaa3', name='install mysql', version=1, action_type='SHELL',
         code='yum install mysql', parameters={}, expected_output=None, expected_rc=None,
         system_kwargs={}, data_mark='20190101000531100000'),
    dict(id='aaaaaaaa-1234-5678-1234-56781234aaa4', name='start mysql', version=1, action_type='SHELL',
         code='service start mysql', parameters={}, expected_output=None, expected_rc=None,
         system_kwargs={}, data_mark='20190101000531100000')
],
    'StepRepo': [
        dict(id='eeeeeeee-1234-5678-1234-56781234eee1', undo=False, stop_on_error=True,
             action_template='aaaaaaaa-1234-5678-1234-56781234aaa1',
             step_expected_output=None, step_expected_rc=0, step_parameters={'dir': 'folder'}, step_system_kwargs=None,
             data_mark='20190101000530100000'),
        dict(id='eeeeeeee-1234-5678-1234-56781234eee2', undo=True, stop_on_error=True,
             action_template='aaaaaaaa-1234-5678-1234-56781234aaa2',
             step_expected_output=None, step_expected_rc=0, step_parameters={'dir': 'folder'},
             step_system_kwargs=None, data_mark='20190101000530100000'),
        dict(id='eeeeeeee-1234-5678-1234-56781234eee3', undo=False, stop_on_error=True,
             action_template='aaaaaaaa-1234-5678-1234-56781234aaa3',
             step_expected_output=None, step_expected_rc=0, step_parameters={}, step_system_kwargs=None,
             data_mark='20190101000532100000'),
        dict(id='eeeeeeee-1234-5678-1234-56781234eee4', undo=True, stop_on_error=True,
             action_template='aaaaaaaa-1234-5678-1234-56781234aaa4',
             step_expected_output=None, step_expected_rc=0, step_parameters={}, step_system_kwargs=None,
             data_mark='20190101000533100000')
    ],
    'OrchestrationRepo': [
        dict(id='cccccccc-1234-5678-1234-56781234ccc1', name='create folder', version=1,
             description='Creates a folder on the specified location',
             steps=['eeeeeeee-1234-5678-1234-56781234eee1',
                    'eeeeeeee-1234-5678-1234-56781234eee2'],
             dependencies={'eeeeeeee-1234-5678-1234-56781234eee1': ['eeeeeeee-1234-5678-1234-56781234eee2'],
                           'eeeeeeee-1234-5678-1234-56781234eee2': []},
             data_mark='20190101000530100000'),
        dict(id='cccccccc-1234-5678-1234-56781234ccc2', name='create folder', version=1,
             description='Creates a folder on the specified location',
             steps=['eeeeeeee-1234-5678-1234-56781234eee3',
                    'eeeeeeee-1234-5678-1234-56781234eee4'],
             dependencies={'eeeeeeee-1234-5678-1234-56781234eee3': ['eeeeeeee-1234-5678-1234-56781234eee4'],
                           'eeeeeeee-1234-5678-1234-56781234eee4': []},
             data_mark='20190101000534100000')
    ],
    'ServerRepo': [
        dict(id='bbbbbbbb-1234-5678-1234-56781234bbb1', name='server1.localdomain', ip='127.0.0.1', port=80, birth=None,
             keep_alive=None, available=True, granules=[], route=[], alt_route=[], data_mark='20190101000530100000'),
        dict(id='bbbbbbbb-1234-5678-1234-56781234bbb2', name='server2.localdomain', ip='127.0.0.1', port=81, birth=None,
             keep_alive=None, available=True, granules=[], route=[], alt_route=[], data_mark='20190101000530100000'),
        dict(id='bbbbbbbb-1234-5678-1234-56781234bbb3', name='server3.localdomain', ip='127.0.0.1', port=80, birth=None,
             keep_alive=None, available=True, granules=[], route=[], alt_route=[], data_mark='20190101000530200000'),
    ],
    'CatalogRepo': [
        dict(entity='ActionTemplate', data_mark='20190101000532100000'),
        dict(entity='Step', data_mark='20190101000533100000'),
        dict(entity='Orchestration', data_mark='20190101000534100000'),
        dict(entity='Server', data_mark='20190101000530200000'),
    ]
}

delta = {'ActionTemplateRepo': [
    dict(id='aaaaaaaa-1234-5678-1234-56781234aaa1', name='mkdir', version=1, action_type='SHELL',
         code='mkdir {dir}', parameters={}, expected_output=None, expected_rc=None,
         system_kwargs={}, data_mark='20190101000532100000'),
    dict(id='aaaaaaaa-1234-5678-1234-56781234aaa3', name='install mysql', version=1, action_type='SHELL',
         code='yum install mysql', parameters={}, expected_output=None, expected_rc=None,
         system_kwargs={}, data_mark='20190101000531100000'),
    dict(id='aaaaaaaa-1234-5678-1234-56781234aaa4', name='start mysql', version=1, action_type='SHELL',
         code='service start mysql', parameters={}, expected_output=None, expected_rc=None,
         system_kwargs={}, data_mark='20190101000531100000')
],
    'StepRepo': [
        dict(id='eeeeeeee-1234-5678-1234-56781234eee3', undo=False, stop_on_error=True,
             action_template='aaaaaaaa-1234-5678-1234-56781234aaa3',
             step_expected_output=None, step_expected_rc=0, step_parameters={}, step_system_kwargs=None,
             data_mark='20190101000532100000'),
        dict(id='eeeeeeee-1234-5678-1234-56781234eee4', undo=True, stop_on_error=True,
             action_template='aaaaaaaa-1234-5678-1234-56781234aaa4',
             step_expected_output=None, step_expected_rc=0, step_parameters={}, step_system_kwargs=None,
             data_mark='20190101000533100000')
    ],
    'OrchestrationRepo': [
        dict(id='cccccccc-1234-5678-1234-56781234ccc2', name='create folder', version=1,
             description='Creates a folder on the specified location',
             steps=['eeeeeeee-1234-5678-1234-56781234eee3',
                    'eeeeeeee-1234-5678-1234-56781234eee4'],
             dependencies={'eeeeeeee-1234-5678-1234-56781234eee3': ['eeeeeeee-1234-5678-1234-56781234eee4'],
                           'eeeeeeee-1234-5678-1234-56781234eee4': []},
             data_mark='20190101000534100000')
    ],
    'ServerRepo': [
        dict(id='bbbbbbbb-1234-5678-1234-56781234bbb3', name='server3.localdomain', ip='127.0.0.1', port=80, birth=None,
             keep_alive=None, available=True, granules=[], route=[], alt_route=[], data_mark='20190101000530200000'),
    ]
}


class Server1(Config):
    DEVELOPMENT = True
    DEBUG = True
    ENV = 'development'
    SERVER_NAME = 'server1.localdomain'
    REPOSITORY = 'memory'
    DM_DATABASE_CONTENT = data1
    # SQLALCHEMY_DATABASE_URI = os.environ.get('DEV_DATABASE_URL') or \
    #                           'sqlite:///' + os.path.join(basedir, 'data-dev.sqlite')
    SQLALCHEMY_DATABASE_URI = None


class Server2(Config):
    DEVELOPMENT = True
    DEBUG = True
    ENV = 'development'
    SERVER_NAME = 'server2.localdomain'
    REPOSITORY = 'memory'
    DM_DATABASE_CONTENT = data2
    # SQLALCHEMY_DATABASE_URI = os.environ.get('DEV_DATABASE_URL') or \
    #                           'sqlite:///' + os.path.join(basedir, 'data-dev.sqlite')
    SQLALCHEMY_DATABASE_URI = None



from unittest import TestCase

import aiohttp
from aiohttp import ClientConnectionError
from aioresponses import aioresponses, CallbackResult
from flask_jwt_extended import create_access_token

from dm.domain.entities import Server, Scope, Locker, State, Route, Catalog
from dm.domain.entities.bootstrap import set_initial
from dm.use_cases.lock import lock_unlock, lock
from dm.web import create_app, db, errors
from dm.web.network import Response


class TestLockUnlock(TestCase):
    def setUp(self):
        """Create and configure a new app instance for each test."""
        # create the app with common test config
        self.app = create_app('test')
        self.app_context = self.app.app_context()
        self.app_context.push()
        self.client = self.app.test_client()
        self.headers = {"Authorization": f"Bearer {create_access_token('00000000-0000-0000-0000-000000000001')}"}

        db.create_all()
        set_initial()

        self.n1 = Server("node1", port=8000)
        Route(self.n1, cost=0)
        self.n2 = Server("node2", port=8000)
        Route(self.n2, cost=0)
        db.session.add_all([self.n1, self.n2])
        db.session.commit()
        self.datamark = Catalog.max_catalog(str)

    def tearDown(self) -> None:
        db.session.remove()
        db.drop_all()
        self.app_context.pop()

    @aioresponses()
    def test_lock_unlock_lock(self, m):
        def callback_prevent(url, **kwargs):
            self.assertDictEqual(kwargs['json'],
                                 {'scope': 'ORCHESTRATION', 'datemark': self.datamark,
                                  'applicant': [str(self.n1.id), str(self.n2.id)]})
            return CallbackResult("{'message': 'Preventing lock acquired'}", status=200)

        def callback_lock(url, **kwargs):
            self.assertDictEqual(kwargs['json'],
                                 {'scope': 'ORCHESTRATION', 'applicant': [str(self.n1.id), str(self.n2.id)]})
            return CallbackResult("{'message': 'Locked'}", status=200)

        m.post(self.n1.url('api_1_0.locker_prevent'), callback=callback_prevent)
        m.post(self.n2.url('api_1_0.locker_prevent'), callback=callback_prevent)
        m.post(self.n1.url('api_1_0.locker_lock'), callback=callback_lock)
        m.post(self.n2.url('api_1_0.locker_lock'), callback=callback_lock)

        ret = lock_unlock('L', Scope.ORCHESTRATION, [self.n1, self.n2], applicant=[str(self.n1.id), str(self.n2.id)])

        self.assertIsNone(ret)

    @aioresponses()
    def test_lock_unlock_lock_with_error_on_preventing(self, m):
        m.post(self.n1.url('api_1_0.locker_prevent'), status=200, payload={'message': 'Preventing lock acquired'})
        m.post(self.n2.url('api_1_0.locker_prevent'), status=409, payload={'error': 'Unable to request for lock'})

        with self.assertRaises(errors.LockerError) as e:
            ret = lock_unlock('L', Scope.ORCHESTRATION, [self.n1, self.n2])

        self.assertEqual(Scope.ORCHESTRATION, e.exception.scope)
        self.assertEqual('prevent', e.exception.action)
        self.assertEqual(1, len(e.exception.responses))
        self.assertListEqual([Response(msg={'error': 'Unable to request for lock'}, code=409, server=self.n2,
                                       url=self.n2.url('api_1_0.locker_prevent'))],
                             e.exception.responses)

    @aioresponses()
    def test_lock_unlock_lock_with_server_error_on_preventing(self, m):
        m.post(self.n1.url('api_1_0.locker_prevent'), status=200, payload={'message': 'Preventing lock acquired'})
        m.post(self.n2.url('api_1_0.locker_prevent'), status=500, body="Error message")

        with self.assertRaises(errors.LockError) as e:
            ret = lock_unlock('L', Scope.ORCHESTRATION, [self.n1, self.n2])

        self.assertEqual(Scope.ORCHESTRATION, e.exception.scope)
        self.assertListEqual(
            [Response("Error message", code=500, server=self.n2, url=self.n2.url('api_1_0.locker_prevent'))],
            e.exception.responses)

    @aioresponses()
    def test_lock_unlock_lock_with_connection_error(self, m):
        m.post(self.n1.url('api_1_0.locker_prevent'), status=200, payload={'message': 'Preventing lock acquired'})
        m.post(self.n2.url('api_1_0.locker_prevent'), exception=aiohttp.ClientConnectionError('test'))

        with self.assertRaises(errors.LockError) as e:
            ret = lock_unlock('L', Scope.ORCHESTRATION, [self.n1, self.n2])

        self.assertEqual(Scope.ORCHESTRATION, e.exception.scope)
        self.assertEqual(1, len(e.exception.responses))
        response = e.exception.responses[0]
        self.assertIsNone(response.msg)
        self.assertIsNone(response.code)
        self.assertEqual(self.n2, response.server)
        self.assertIsInstance(response.exception, aiohttp.ClientConnectionError)
        self.assertTupleEqual(('test',), response.exception.args)

    @aioresponses()
    def test_lock_unlock_lock_error_on_lock(self, m):
        m.post(self.n1.url('api_1_0.locker_prevent'), status=200, payload={'message': 'Preventing lock acquired'})
        m.post(self.n2.url('api_1_0.locker_prevent'), status=200, payload={'message': 'Preventing lock acquired'})
        m.post(self.n1.url('api_1_0.locker_lock'), status=200, payload={'message': 'Locked'})
        m.post(self.n2.url('api_1_0.locker_lock'), status=409, payload={'error': 'Unable to lock'})

        with self.assertRaises(errors.LockError) as e:
            ret = lock_unlock('L', Scope.ORCHESTRATION, [self.n1, self.n2])

        self.assertEqual(Scope.ORCHESTRATION, e.exception.scope)
        self.assertListEqual([Response(msg={'error': 'Unable to lock'}, code=409, server=self.n2,
                                       url=self.n2.url('api_1_0.locker_lock'))],
                             e.exception.responses)

    @aioresponses()
    def test_lock_unlock_unlock(self, m):
        def callback_unlock(url, **kwargs):
            self.assertDictEqual(kwargs['json'],
                                 {'scope': 'ORCHESTRATION', 'applicant': [str(self.n1.id), str(self.n2.id)]})
            return CallbackResult(payload={'message': 'UnLocked'}, status=200)

        m.post(self.n1.url('api_1_0.locker_unlock'), callback=callback_unlock)
        m.post(self.n2.url('api_1_0.locker_unlock'), callback=callback_unlock)

        ret = lock_unlock('U', Scope.ORCHESTRATION, [self.n1, self.n2])

        self.assertIsNone(ret)

    @aioresponses()
    def test_lock_unlock_unlock_with_error(self, m):
        m.post(self.n2.url('api_1_0.locker_unlock'), status=200, payload={'message': 'UnLocked'})
        m.post(self.n1.url('api_1_0.locker_unlock'), status=409, payload={'error': 'Unable to unlock.'})

        with self.assertRaises(errors.LockError) as e:
            ret = lock_unlock('U', Scope.ORCHESTRATION, [self.n1, self.n2], [str(self.n1.id), str(self.n2.id)])

        self.assertEqual(Scope.ORCHESTRATION, e.exception.scope)
        self.assertListEqual([Response(msg={'error': 'Unable to unlock.'}, code=409, server=self.n1,
                                       url=self.n1.url('api_1_0.locker_unlock'))],
                             e.exception.responses)


class TestLock(TestCase):
    def setUp(self):
        """Create and configure a new app instance for each test."""
        # create the app with common test config
        self.app = create_app('test')
        self.app_context = self.app.app_context()
        self.app_context.push()
        self.client = self.app.test_client()
        self.headers = {"Authorization": f"Bearer {create_access_token('00000000-0000-0000-0000-000000000001')}"}

        db.create_all()
        set_initial()

        self.n1 = Server("node1", port=8000)
        Route(self.n1, cost=0)
        self.n2 = Server("node2", port=8000)
        Route(self.n2, cost=0)
        db.session.add_all([self.n1, self.n2])
        db.session.commit()
        self.datemark = Catalog.max_catalog(str)

    def tearDown(self) -> None:
        db.session.remove()
        db.drop_all()
        self.app_context.pop()

    def test_lock_no_server(self):
        with self.assertRaises(RuntimeError):
            ret = lock(Scope.ORCHESTRATION)

    @aioresponses()
    def test_lock_catalog(self, m):
        def callback_prevent(url, **kwargs):
            self.assertDictEqual(kwargs['json'],
                                 {'scope': 'CATALOG', 'datemark': self.datemark,
                                  'applicant': [str(Server.get_current().id), str(self.n1.id),
                                                str(self.n2.id)]})
            return CallbackResult("{'message': 'Preventing lock acquired'}", status=200)

        def callback_lock(url, **kwargs):
            self.assertDictEqual(kwargs['json'],
                                 {'scope': 'CATALOG', 'applicant': [str(Server.get_current().id), str(self.n1.id),
                                                                    str(self.n2.id)]})
            return CallbackResult("{'message': 'Locked'}", status=200)

        m.post(Server.get_current().url('api_1_0.locker_prevent'), callback=callback_prevent)
        m.post(self.n1.url('api_1_0.locker_prevent'), callback=callback_prevent)
        m.post(self.n2.url('api_1_0.locker_prevent'), callback=callback_prevent)
        m.post(Server.get_current().url('api_1_0.locker_lock'), callback=callback_lock)
        m.post(self.n1.url('api_1_0.locker_lock'), callback=callback_lock)
        m.post(self.n2.url('api_1_0.locker_lock'), callback=callback_lock)

        applicant = lock(Scope.CATALOG)

        self.assertEqual(applicant, [str(Server.get_current().id), str(self.n1.id), str(self.n2.id)])

    @aioresponses()
    def test_lock_catalog_error_on_preventing(self, m):

        def callback_prevent(url, **kwargs):
            self.assertDictEqual(kwargs['json'],
                                 {'scope': 'CATALOG', 'applicant': [str(Server.get_current().id), str(self.n1.id),
                                                                    str(self.n2.id)]})
            return CallbackResult("{'message': 'Preventing lock acquired'}", status=200)

        def callback_unlock(url, **kwargs):
            self.assertDictEqual(kwargs['json'], {'scope': 'CATALOG', 'action': 'UNLOCK',
                                                  'applicant': [str(Server.get_current().id), str(self.n1.id),
                                                                str(self.n2.id)]})
            return CallbackResult("{'message': 'UnLocked'}", status=200)

        m.post(Server.get_current().url('api_1_0.locker_prevent'), callback=callback_prevent)
        m.post(self.n1.url('api_1_0.locker_prevent'), exception=ClientConnectionError())
        m.post(self.n2.url('api_1_0.locker_prevent'), callback=callback_prevent)
        m.post(Server.get_current().url('api_1_0.locker_unlock'), callback=callback_unlock)
        m.post(self.n2.url('api_1_0.locker_unlock'), callback=callback_unlock)

        with self.assertRaises(errors.LockError):
            applicant = lock(Scope.CATALOG)

        c = Locker.query.get(Scope.CATALOG)
        self.assertEqual(State.UNLOCKED, c.state)

    @aioresponses()
    def test_lock_unreachable_network(self, m):
        def callback_prevent(url, **kwargs):
            return CallbackResult("{'message': 'Preventing lock acquired'}", status=200)

        def callback_unlock(url, **kwargs):
            return CallbackResult("{'message': 'UnLocked'}", status=200)

        m.post(self.n1.url('api_1_0.locker_prevent'), callback=callback_prevent)
        m.post(self.n2.url('api_1_0.locker_prevent'), callback=callback_prevent)
        m.post(self.n2.url('api_1_0.locker_unlock'), callback=callback_unlock)

        self.n1.route.cost = None
        self.n1.route.gate = None
        self.n1.route.proxy_server = None

        with self.assertRaises(errors.LockError) as e:
            applicant = lock(Scope.CATALOG, servers=[self.n1, self.n2])

        self.assertEqual(Scope.CATALOG, e.exception.scope)
        self.assertEqual(errors.LockError.action_map['P'], e.exception.action)
        self.assertDictEqual(errors.format_error_content(errors.LockError(Scope.CATALOG, action='P', responses=[
            Response(exception=errors.UnreachableDestination(self.n1), server=self.n1)])),
                             errors.format_error_content(e.exception))
        self.assertEqual(2, len(m.requests))



import base64
import os

from flask import request, current_app
from flask_jwt_extended import jwt_required
from flask_restful import Resource

from dm.domain.entities import Log, Server
from dm.domain.entities.log import Mode
from dm.utils.helpers import clean_string
from dm.web import db, errors
from dm.web.decorators import forward_or_dispatch, securizer, validate_schema, lock_catalog
from dm.web.helpers import filter_query, check_param_in_uri
from dm.web.json_schemas import log_post, logs_post, log_patch


class LogList(Resource):

    @securizer
    @jwt_required
    @forward_or_dispatch
    def get(self):
        query = filter_query(Log, request.args)
        return [log.to_json(human=check_param_in_uri('human'), delete_data=False) for log in query.all()]

    @forward_or_dispatch
    @jwt_required
    @securizer
    @validate_schema(logs_post)
    @lock_catalog
    def post(self):
        data = request.get_json()
        source_server = Server.query.get_or_404(data.pop('src_server_id'))
        destination_server = Server.query.get_or_404(data.pop('dst_server_id'))
        if source_server == destination_server:
            return {'error': 'source and destination must be different'}, 400
        if 'mode' in data:
            data['mode'] = Mode[data['mode']]
            if data['mode'] == Mode.FOLDER:
                if 'dest_folder' not in data:
                    raise errors.ParameterMustBeSet("property 'dest_folder' must be set when mode=FOLDER")
        log = Log(source_server=source_server, destination_server=destination_server, **data)
        db.session.add(log)
        db.session.commit()
        return {'id': str(log.id)}, 201


class LogResource(Resource):

    @securizer
    @jwt_required
    @forward_or_dispatch
    def get(self, log_id):
        return Log.query.get_or_404(log_id).to_json()

    @securizer
    @jwt_required
    @forward_or_dispatch
    @validate_schema(log_post)
    def post(self, log_id):
        log = Log.query.get_or_404(log_id)
        data = request.get_json()
        file = data.get('file')
        if log.mode in (Mode.REPO_MIRROR, Mode.REPO_ROOT):
            file = file.format(
                LOG_REPO=os.path.join(current_app.config['LOG_REPO'], clean_string(log.source_server.name)))
        data_log = base64.b64decode(data.get('data').encode('ascii'))
        try:
            if not os.path.exists(os.path.dirname(file)):
                os.makedirs(os.path.dirname(file))
            with open(file, 'ab') as fh:
                fh.write(data_log)
        except Exception as e:
            return {"error": str(e)}
        return {'offset': os.path.getsize(file)}

    @securizer
    @jwt_required
    @forward_or_dispatch
    @validate_schema(log_patch)
    @lock_catalog
    def patch(self, log_id):
        log = Log.query.get_or_404(log_id)
        data = request.get_json()
        if 'include' in data and log.include != data.get('include'):
            log.include = data.get('include')
        if 'exclude' in data and log.exclude != (data.get('exclude') or '^$'):
            log.exclude = data.get('exclude')
        if 'recursive' in data and log.recursive != data.get('recursive'):
            log.recursive = data.get('recursive')
        if 'mode' in data and log.mode != data.get('mode'):
            log.mode = data.get('mode')
        if log.mode == Mode.FOLDER:
            if 'dest_folder' in data and log.dest_folder != data.get('dest_folder'):
                log.dest_folder = data.get('dest_folder')
            if log.dest_folder == None:
                raise errors.ParameterMustBeSet("property 'dest_folder' must be set when mode=FOLDER")
        if log in db.session.dirty:
            db.session.commit()
            return {}, 204
        return {}, 202

    @securizer
    @jwt_required
    @forward_or_dispatch
    @lock_catalog
    def delete(self, log_id):
        log = Log.query.get_or_404(log_id)
        log.delete()
        db.session.commit()
        return {}, 204



import typing as t
from datetime import datetime

from dm import defaults
from dm.utils.typos import UtcDateTime
from dm.web import db


class Catalog(db.Model):
    __tablename__ = 'L_catalog'
    entity = db.Column(db.String(40), primary_key=True, unique=True)
    last_modified_at = db.Column(UtcDateTime(timezone=True), nullable=False)

    def __init__(self, entity: str, last_modified_at: datetime):
        self.entity = entity
        self.last_modified_at = last_modified_at

    def __repr__(self):
        return f'<{self.__class__.__name__}({self.entity}, {self.last_modified_at})>'

    @staticmethod
    def max_catalog(out=None) -> t.Union[datetime, str]:
        catalog_ver = db.session.query(db.func.max(Catalog.last_modified_at)).scalar()
        if catalog_ver is None:
            catalog_ver = defaults.INITIAL_DATEMARK
        return catalog_ver.strftime(defaults.DATEMARK_FORMAT) if out is str else catalog_ver



from flask import Blueprint
from flask_restful import Api


api_bp = Blueprint('api_1_0', __name__, url_prefix='/api/v1.0')
api = Api(api_bp)

# import routes
import dm.web.api_1_0.urls.locker
import dm.web.api_1_0.resources.transfer
import dm.web.api_1_0.urls.use_cases

from dm.web.api_1_0.resources import *

# generate resources Flask_Restful
api.add_resource(ActionTemplateList, '/action_templates')
api.add_resource(ActionTemplateResource, '/action_templates/<action_template_id>')

api.add_resource(LogList, '/log')
api.add_resource(LogResource, '/log/<log_id>')

api.add_resource(OrchestrationList, '/orchestrations')
api.add_resource(OrchestrationResource, '/orchestrations/<orchestration_id>')
api.add_resource(OrchestrationExecutionRelationship, '/orchestrations/<orchestration_id>/executions')

api.add_resource(OrchExecutionList, '/orchestration_executions')
api.add_resource(OrchExecutionResource, '/orchestration_executions/<execution_id>')
api.add_resource(OrchExecStepExecRelationship, '/orchestration_executions/<execution_id>/step_executions')

api.add_resource(ServerList, '/servers')
api.add_resource(ServerResource, '/servers/<server_id>')
api.add_resource(GranuleList, '/granules')

api.add_resource(SoftwareList, '/software')
api.add_resource(SoftwareResource, '/software/<software_id>')
api.add_resource(SoftwareServersResource, '/software/<software_id>/servers')

api.add_resource(StepList, '/steps')
api.add_resource(StepResource, '/steps/<step_id>')
api.add_resource(StepRelationshipParents, '/steps/<step_id>/relationship/parents')
api.add_resource(StepRelationshipChildren, '/steps/<step_id>/relationship/children')

api.add_resource(StepExecutionList, '/step_executions')
api.add_resource(StepExecutionResource, '/step_executions/<execution_id>')

api.add_resource(TransferList, '/transfers')
api.add_resource(TransferResource, '/transfers/<transfer_id>')

api.add_resource(UserList, '/users')
api.add_resource(UserResource, '/users/<user_id>')


from flask import request
from flask_jwt_extended import jwt_required
from flask_restful import Resource

from dm.domain.entities import ActionTemplate, ActionType
from dm.web import db
from dm.web.decorators import securizer, forward_or_dispatch, validate_schema, lock_catalog
from dm.web.helpers import filter_query
from dm.web.json_schemas import action_template_patch, action_template_post


class ActionTemplateList(Resource):

    @forward_or_dispatch
    @jwt_required
    @securizer
    def get(self):
        query = filter_query(ActionTemplate, request.args)
        return [at.to_json() for at in query.all()]

    @forward_or_dispatch
    @jwt_required
    @securizer
    @validate_schema(action_template_post)
    @lock_catalog
    def post(self):
        json_data = request.get_json()
        if isinstance(json_data, dict):
            json_data = [json_data]
        resp_data = []
        for json_at in json_data:
            json_at['action_type'] = ActionType[json_at['action_type']]
            if 'version' not in json_at:
                json_at['version'] = ActionTemplate.query.filter_by(name=json_at['name']).count() + 1
            at = ActionTemplate(**json_at)
            db.session.add(at)
            data = {'id': at.id}
            if 'version' not in json_at:
                data.update(version=at.version)
            resp_data.append(data)
        db.session.commit()
        return resp_data[0] if isinstance(request.get_json(), dict) else resp_data, 201


class ActionTemplateResource(Resource):
    @forward_or_dispatch
    @jwt_required
    @securizer
    def get(self, action_template_id):
        return ActionTemplate.query.get_or_404(action_template_id).to_json()

    @securizer
    @jwt_required
    @forward_or_dispatch
    @validate_schema(action_template_patch)
    @lock_catalog
    def patch(self, action_template_id):
        at = ActionTemplate.query.get_or_404(action_template_id)
        data = request.get_json()
        if 'action_type' in data and at.action_type != ActionType[data.get('action_type')]:
            at.action_type = ActionType[data.get('action_type')]
        if 'code' in data and at.code != (data.get('code')):
            at.code = data.get('code')
        if 'parameters' in data and at.parameters != data.get('parameters'):
            at.parameters = data.get('parameters')
        if 'expected_stdout' in data and at.expected_stdout != data.get('expected_stdout'):
            at.expected_stdout = data.get('expected_stdout')
        if 'expected_stderr' in data and at.expected_stderr != data.get('expected_stderr'):
            at.expected_stderr = data.get('expected_stderr')
        if 'expected_rc' in data and at.expected_rc != data.get('expected_rc'):
            at.expected_rc = data.get('expected_rc')
        if 'pre_process' in data and at.pre_process != data.get('pre_process'):
            at.pre_process = data.get('pre_process')
        if 'post_process' in data and at.post_process != data.get('post_process'):
            at.post_process = data.get('post_process')
        if at in db.session.dirty:
            db.session.commit()
            return {}, 204
        return {}, 202



import itertools
import os
import socket
from datetime import datetime, timezone

import netifaces

DATETIME_FORMAT = "%m/%d/%Y, %H:%M:%S.%f %z"
DATEMARK_FORMAT = "%Y%m%d.%H%M%S.%f%z"
HOME = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
HOSTNAME = socket.gethostname()

ips = list(filter(lambda x: x != '127.0.0.1', itertools.chain(
    *[[ip['addr'] for ip in netifaces.ifaddresses(iface).get(netifaces.AF_INET, [])] for iface in
      netifaces.interfaces()])))

MIN_SERVERS_QUORUM = 5  # minimum servers to run quorum algorithm
INITIAL_DATEMARK = datetime(2019, 4, 1, tzinfo=timezone.utc)
DEFAULT_PORT = 5000
LOOPBACK_PORT = 20194

MAX_WAITING_TIME = 300  # time in seconds waiting tasks to finish
MAX_TIME_WAITING_SERVERS = 600  # max time waiting for servers to be created

TIMEOUT_REQUEST = 60
TIMEOUT_PREVENTING_LOCK = 60  # max time in seconds locker will be in PREVENTING_LOCK before returning to UNLOCK
TIMEOUT_ORCHESTRATION = 1800  # max time waiting for an orchestration to finish
TIMEOUT_COMMAND = 20  # max time waiting for a command execution

CHUNK_SIZE = 2*1024  # in MB
MAX_SENDERS = 4



import re
from unittest import TestCase

from aioresponses import aioresponses, CallbackResult
from flask_jwt_extended import create_access_token

from dm.domain.entities import Server, Scope, Locker, State, Dimension, Route
from dm.domain.entities.bootstrap import set_initial
from dm.use_cases.lock import lock_scope
from dm.utils.helpers import generate_dimension
from dm.web import create_app, db, load_global_data_into_context


class TestLockScope(TestCase):
    def setUp(self):
        """Create and configure a new app instance for each test."""
        # create the app with common test config
        self.app = create_app('test')
        self.app_context = self.app.app_context()
        self.app_context.push()
        self.client = self.app.test_client()
        self.headers = {"Authorization": f"Bearer {create_access_token('00000000-0000-0000-0000-000000000001')}"}

        db.create_all()
        set_initial()
        db.session.commit()

    def tearDown(self) -> None:
        db.session.remove()
        db.drop_all()
        self.app_context.pop()

    @aioresponses()
    def test_lock_scope(self, m):
        n1 = Server("node1", port=8000)
        Route(n1, cost=0)
        n2 = Server("node2", port=8000)
        Route(n2, cost=0)
        db.session.add_all([n1, n2])
        db.session.commit()

        def callback_prevent(url, **kwargs):
            return CallbackResult("{'message': 'Preventing lock acquired'}", status=200)

        def callback_lock(url, **kwargs):
            return CallbackResult("{'message': 'Locked'}", status=200)

        def callback_unlock(url, **kwargs):
            return CallbackResult("{'message': 'UnLocked'}", status=200)

        def callback_client(url, **kwargs):
            kwargs.pop('allow_redirects')
            # workarround for https://github.com/pnuckowski/aioresponses/issues/111
            headers = {'Authorization': f"Bearer {create_access_token('00000000-0000-0000-0000-000000000001')}"}

            r = self.client.post(url.path, json=kwargs['json'], headers=headers)

            return CallbackResult(r.data, status=r.status_code)

        m.post(re.compile(Server.get_current().url() + '.*'), callback=callback_client)
        m.post(n1.url('api_1_0.locker_prevent'), callback=callback_prevent)
        m.post(n2.url('api_1_0.locker_prevent'), callback=callback_prevent)
        m.post(re.compile(Server.get_current().url() + '.*'), callback=callback_client)
        m.post(n1.url('api_1_0.locker_lock'), callback=callback_lock)
        m.post(n2.url('api_1_0.locker_lock'), callback=callback_lock)
        m.post(re.compile(Server.get_current().url() + '.*'), callback=callback_client)
        m.post(n1.url('api_1_0.locker_unlock'), callback=callback_unlock)
        m.post(n2.url('api_1_0.locker_unlock'), callback=callback_unlock)

        l = Locker.query.get(Scope.CATALOG)
        self.assertEqual(State.UNLOCKED, l.state)
        self.assertEqual(None, l.applicant)

        with self.app.test_request_context('/api/v1.0/lock'):
            with lock_scope(Scope.CATALOG):
                l = Locker.query.get(Scope.CATALOG)
                self.assertEqual(State.LOCKED, l.state)
                self.assertEqual([str(Server.get_current().id), str(n1.id), str(n2.id)], l.applicant)

        l = Locker.query.get(Scope.CATALOG)
        self.assertEqual(State.UNLOCKED, l.state)
        self.assertEqual(None, l.applicant)

    @aioresponses()
    def test_lock_scope_packing(self, m):
        self.app.config['SECURIZER'] = True
        dim = generate_dimension('dimension')
        dim.current = True
        db.session.add(dim)
        n1 = Server("node1", port=8000)
        Route(n1, cost=0)
        n2 = Server("node2", port=8000)
        Route(n2, cost=0)
        db.session.add_all([n1, n2])
        db.session.commit()

        def callback_prevent(url, **kwargs):
            return CallbackResult("{'message': 'Preventing lock acquired'}", status=200)

        def callback_lock(url, **kwargs):
            return CallbackResult("{'message': 'Locked'}", status=200)

        def callback_unlock(url, **kwargs):
            return CallbackResult("{'message': 'UnLocked'}", status=200)

        def callback_client(url, **kwargs):
            kwargs.pop('allow_redirects')
            # workarround for https://github.com/pnuckowski/aioresponses/issues/111
            headers = {'Authorization': f"Bearer {create_access_token('00000000-0000-0000-0000-000000000001')}"}

            r = self.client.post(url.path, json=kwargs['json'], headers=headers)

            return CallbackResult(r.data, status=r.status_code)

        m.post(re.compile(Server.get_current().url()+'.*'), callback=callback_client)
        m.post(n1.url('api_1_0.locker_prevent'), callback=callback_prevent)
        m.post(n2.url('api_1_0.locker_prevent'), callback=callback_prevent)
        m.post(re.compile(Server.get_current().url()+'.*'), callback=callback_client)
        m.post(n1.url('api_1_0.locker_lock'), callback=callback_lock)
        m.post(n2.url('api_1_0.locker_lock'), callback=callback_lock)
        m.post(re.compile(Server.get_current().url()+'.*'), callback=callback_client)
        m.post(n1.url('api_1_0.locker_unlock'), callback=callback_unlock)
        m.post(n2.url('api_1_0.locker_unlock'), callback=callback_unlock)

        l = Locker.query.get(Scope.CATALOG)
        self.assertEqual(State.UNLOCKED, l.state)
        self.assertEqual(None, l.applicant)

        with self.app.test_request_context('/api/v1.0/lock'):
            load_global_data_into_context()
            with lock_scope(Scope.CATALOG):
                l = Locker.query.get(Scope.CATALOG)
                self.assertEqual(State.LOCKED, l.state)
                self.assertEqual([str(Server.get_current().id), str(n1.id), str(n2.id)], l.applicant)

        l = Locker.query.get(Scope.CATALOG)
        self.assertEqual(State.UNLOCKED, l.state)
        self.assertEqual(None, l.applicant)



class TestLockScopeFullChain(TestCase):
    def setUp(self):
        """Create and configure a new app instance for each test."""
        # create the app with common test config
        self.app1 = create_app('test')
        self.app1.config['SERVER_NAME'] = 'node1'
        self.app1.config['SECURIZER'] = True
        self.client1 = self.app1.test_client()
        self.app2 = create_app('test')
        self.app2.config['SERVER_NAME'] = 'node2'
        self.app2.config['SECURIZER'] = True
        self.client2 = self.app2.test_client()

        with self.app1.app_context():
            db.create_all()
            set_initial()
            s = Server.get_current()
            s.gates = []
            s.add_new_gate('node1', 8000)
            dim = generate_dimension('dimension')
            dim.current = True
            db.session.add(dim)
            db.session.commit()
            self.s1_json = Server.get_current().to_json()
            self.dim_json = dim.to_json()
            self.headers = {"Authorization": f"Bearer {create_access_token('00000000-0000-0000-0000-000000000001')}"}

        with self.app2.app_context():
            db.create_all()
            set_initial()
            s = Server.get_current()
            s.gates = []
            s.add_new_gate('node2', 8000)
            db.session.commit()
            self.s2_json = Server.get_current().to_json()
            s = Server.from_json(self.s1_json)
            s.add_new_gate('node1', 8000)
            Route(s, cost=0)
            db.session.add(s)
            dim = Dimension.from_json(self.dim_json)
            dim.current = True
            db.session.add(dim)
            db.session.commit()

        with self.app1.app_context():
            s = Server.from_json(self.s2_json)
            s.add_new_gate('node2', 8000)
            Route(s, cost=0)
            db.session.add(s)
            db.session.commit()

    def tearDown(self) -> None:
        with self.app1.app_context():
            db.session.remove()
            db.drop_all()

        with self.app2.app_context():
            db.session.remove()
            db.drop_all()

    @aioresponses()
    def test_lock_catalog(self, m):
        def callback_client1(url, **kwargs):
            kwargs.pop('allow_redirects')
            # passing headers as a workarround for https://github.com/pnuckowski/aioresponses/issues/111
            r = self.client1.post(url.path, json=kwargs['json'], headers=self.headers)

            return CallbackResult(r.data, status=r.status_code)

        def callback_client2(url, **kwargs):
            kwargs.pop('allow_redirects')
            # passing headers as a workarround for https://github.com/pnuckowski/aioresponses/issues/111
            r = self.client2.post(url.path, json=kwargs['json'], headers=self.headers)

            return CallbackResult(r.data, status=r.status_code)

        m.post(re.compile('https?://node1.*'), callback=callback_client1, repeat=True)
        m.post(re.compile('https?://node2.*'), callback=callback_client2, repeat=True)

        with self.app1.app_context():
            # load dimension into variable
            load_global_data_into_context()
            with self.app1.test_request_context('/api/v1.0/lock'):
                with lock_scope(Scope.CATALOG):
                    l = Locker.query.get(Scope.CATALOG)
                    self.assertEqual(State.LOCKED, l.state)
                    self.assertEqual([self.s1_json['id'], self.s2_json['id']], l.applicant)



import base64
import copy
import json
import pickle
import typing as t

import rsa

from dm.domain.entities import Server
from dm.utils.helpers import encrypt, decrypt
from .exceptions import NotValidMessage

if t.TYPE_CHECKING:
    pass

"""
Singleton class that allows communication between servers. create a routes dict like follows:

--------
Create all servers
>>>s1 = Server(name='Server1', ip='127.0.0.1', port=5001, route=[], id=uuid.UUID('12345678-1234-5678-1234-567812345678'))
>>>s2 = Server(name='Server2', ip='127.0.0.1', port=5002, route=[], id=uuid.UUID('22345678-1234-5678-1234-567812345678'))
>>>s3 = Server(name='Server3', ip='127.0.0.1', port=5003, route=[s2], id=uuid.UUID('32345678-1234-5678-1234-567812345678'))
>>>s4 = Server(name='Server4', ip='127.0.0.1', port=5004, route=[s2, s3], id=uuid.UUID('42345678-1234-5678-1234-567812345678'))
Create dict like servers
>>>dict_route = {s1.id: s1, s2.id: s2, s3.id: s3, s4.id: s4}
>>>g = Gateway(routes=dict_route)
"""



def pack_msg(data,
             destination: t.Union[Server, str] = None,
             source: t.Union[Server, str] = None,
             pub_key: rsa.PublicKey = None,
             priv_key: rsa.PrivateKey = None,
             cipher_key: bytes = None,
             symmetric_key: bytes = None,
             add_key=False) -> t.Dict[str, t.Any]:
    """
    formats data in a well known encrypted structure. See Return.

    Parameters
    ----------
    destination
    source
    pub_key:
        public key object rsa.PublicKey. Used for encrypting the symmetric key
    priv_key:
        private key object rsa.PrivateKey. Private key must be provided for signing structure and when cipher_key
        provided. If not, no signature will be added.
    cipher_key:
        encrypted symmetric key used for data encryption. If set, symmetric_key must be None
    symmetric_key:
        symmetric key to be used for data encryption. If None, randomly generated from cryptography.fernet.Fernet.generate_key()
    data:
        data to encrypt. Data is pickled and then encoded

    Returns
    -------
    returns a dict with the following structure:
        { "destination": "UUID",
          "source": "UUID",
          "enveloped_data": "encrypted_data",
          "key": "encrypted symmetric key",
          "signature": "current data signature"
        }

    if no priv_key specified it does not encrypt the data. 'key' and 'signature' will not append to the structure.
    if symmetric_key or cipher_key given, 'key' will not append to the structure
    """
    try:
        dumped_data: bytes = json.dumps(data).encode('utf-8')
    except TypeError:
        dumped_data: bytes = pickle.dumps(data)

    # get symmetric_key
    if cipher_key:
        if not priv_key:
            raise ValueError('priv_key must be provided to decrypt cipher_key')
        else:
            symmetric_key = rsa.decrypt(cipher_key, priv_key=priv_key)

    # encrypt data
    if pub_key or symmetric_key:
        encrypted_data, new_symmetric_key = encrypt(
            dumped_data,
            symmetric_key=symmetric_key)
    else:
        encrypted_data, new_symmetric_key = dumped_data, None

    msg = dict(enveloped_data=base64.b64encode(encrypted_data).decode('ascii'))

    if destination:
        # warnings.warn("The 'destination' parameter is deprecated, "
        #               "use 'D-Destination header' instead", DeprecationWarning, 2)
        msg.update(destination=str(destination.id) if isinstance(destination, Server) else destination)
    if source:
        msg.update(source=str(source.id) if isinstance(source, Server) else source)
    if pub_key and (new_symmetric_key or add_key):
        msg.update(key=base64.b64encode(rsa.encrypt(new_symmetric_key or symmetric_key, pub_key)).decode('ascii'))
    if priv_key:
        signature = rsa.sign(json.dumps(msg, sort_keys=True).encode('ascii'), priv_key, 'SHA-512')
        msg.update(signature=base64.b64encode(signature).decode('ascii'))
    return msg


def unpack_msg(msg, pub_key: rsa.PublicKey = None, priv_key: rsa.PrivateKey = None, symmetric_key=None,
               cipher_key=None):
    """
    Unpacks msg. If signature in msg, pub_key must be provided to validate it.  Pub_key must be provided to decrypt
    cipher_key if provided.

    Parameters
    ----------
    msg:
        message to be unpacked
    pub_key:
        public key which will be used for signature validation
    priv_key:
        pivate key used for decrypt the cipher_key
    cipher_key:
        encrypted symmetric key used for data encryption. When None checks in the msg if 'key' specified. If Not, then
        symmetric_key may be used.
    symmetric_key:
        symmetric key to be used for data encryption. If None, cipher_key will be used

    Returns
    -------

    """
    if 'enveloped_data' not in msg:
        raise NotValidMessage('msg was not packed')
    if 'signature' in msg:
        if not pub_key:
            raise ValueError('No public key specified to validate signature')
        signature = base64.b64decode(msg.get('signature').encode('ascii'))
        msg_to_validate = copy.deepcopy(msg)
        msg_to_validate.pop('signature')
        rsa.verify(json.dumps(msg_to_validate, sort_keys=True).encode('ascii'), signature, pub_key)

    cipher_key = base64.b64decode(msg.get('key', '').encode('ascii')) or cipher_key
    if cipher_key:
        if not priv_key:
            raise ValueError('No private key specified to decrpyt cipher_key')
        else:
            symmetric_key = rsa.decrypt(cipher_key, priv_key)

    enveloped_data = msg.get('enveloped_data')

    if symmetric_key:
        unloaded_data = decrypt(base64.b64decode(enveloped_data.encode('ascii')), symmetric_key)
    else:
        unloaded_data = base64.b64decode(enveloped_data.encode('ascii'))

    try:
        data = pickle.loads(unloaded_data)
    except pickle.PickleError:
        data = json.loads(unloaded_data)

    return data





import typing as t
from datetime import datetime

import rsa

from dm import defaults
from dm.domain.entities.base import EntityReprMixin, UUIDEntityMixin
from dm.utils.helpers import get_now
from dm.utils.typos import PrivateKey, PublicKey, UtcDateTime
from dm.web import db


class Dimension(db.Model, UUIDEntityMixin, EntityReprMixin):
    __tablename__ = 'L_dimension'
    name = db.Column(db.String(40), nullable=False, unique=True)
    private = db.Column(PrivateKey, unique=True)
    public = db.Column(PublicKey, unique=True)
    current = db.Column(db.Boolean, nullable=False)
    created_at = db.Column(UtcDateTime(timezone=True), default=get_now)

    def __init__(self, name: str, private: t.Union[rsa.PrivateKey, bytes] = None,
                 public: t.Union[rsa.PublicKey, bytes] = None, created_at: datetime = get_now(), current=False,
                 **kwargs):
        UUIDEntityMixin.__init__(self, **kwargs)
        self.name = name
        if isinstance(private, bytes):
            self.private = rsa.PrivateKey.load_pkcs1(private)
        else:
            self.private = private
        if isinstance(public, bytes):
            self.public = rsa.PublicKey.load_pkcs1(public)
        else:
            self.public = public
        self.created_at = created_at
        self.current = current

    @classmethod
    def get_current(cls) -> 'Dimension':
        return db.session.query(cls).filter_by(current=True).one_or_none()

    def to_json(self):
        return {'id': str(self.id) if self.id else None, 'name': self.name,
                'private': self.private.save_pkcs1().decode('ascii'),
                'public': self.public.save_pkcs1().decode('ascii'),
                'created_at': self.created_at.strftime(defaults.DATETIME_FORMAT)}

    @classmethod
    def from_json(cls, kwargs):
        return cls(id=kwargs.get('id'), name=kwargs.get('name'),
                   private=rsa.PrivateKey.load_pkcs1(kwargs.get('private')),
                   public=rsa.PublicKey.load_pkcs1(kwargs.get('public')),
                   created_at=datetime.strptime(kwargs.get('created_at'), defaults.DATETIME_FORMAT)
                   )
